# coding=utf-8
""" Contains objects related to tables.

Classes in the table file help manage the tables
that are used by the application.
"""

# import logging
import re
from abc import ABC as abstract_base
from geoalchemy2 import Geometry
from sqlalchemy import Table as AlchemyTable, Column, MetaData
from src.oogeo.objects.cursor import SearchCursor, InsertCursor, UpdateCursor
from src.oogeo.objects.selectionset import SelectionSet
from src.oogeo.functions.database import build_query_string, build_update_clause, create_insert_row
from src.oogeo.functions.conversions import parse_expression


def load_alchemy_table(workspace, table_name, schema=None):
    """

    :param workspace:
    :param table_name:
    :return:
    """
    # set proper schema
    if not schema:
        schema = workspace.schema

    # reload the alchemy table to include the new field
    geometry_columns = workspace.get_geometry_columns(table_name, schema=schema)
    if geometry_columns:
        columns = [Column(geometry_column['name'], Geometry(geometry_column['type'], srid=geometry_column['srid'])) for geometry_column in geometry_columns]

        return AlchemyTable(table_name, MetaData(), *columns,
            autoload=True, autoload_with=workspace.alchemy_engine, schema=schema)
    else:
        return AlchemyTable(table_name, MetaData(), autoload=True, autoload_with=workspace.alchemy_engine, schema=schema)


class TableBase(abstract_base):
    """ Abstract baseclass for supporting table objects.

    The `TableBase` class should never be initialized directly, but should be accessed through its inheritors,
    the `Table` class and the `Layer` class. The `Table` class acts as an initializer to properly initialize tables
    in ArcPy. The `Layer` class will initialize a layer instance in ArcPy, and also extends the basic table
    functionality with additional methods specific to layer manipulation.

    """

    def __init__(self, table, workspace, schema=None):
        """ Initializes the table base object.

        :param table_name: The name of the table as it is stored in SQLite.
        :param workspace: The workspace where the table is located.
        """
        self.alchemy_table = table if isinstance(table, AlchemyTable) else load_alchemy_table(workspace, table, schema=schema)
        self.table_name = self.alchemy_table.name
        self.schema = self.alchemy_table.schema
        self.deleted = False
        self._cached_oid_field = None
        self.workspace = workspace
        self._selections = []
        self._cached_query_string = None
        self._cached_update_clause = None


    def __enter__(self):
        """ Initialized the table if necessary and returns a reference to itself.

        :return: self
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Cleans up table files when exiting a `with` statement.

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.delete()

    # SELECTION METHODS
    def select_layer_by_attribute(self, query_string, selection_type='NEW_SELECTION'):
        """ Creates a selection set in the table based on an attribute query.

        This method selects records in the table based on the query string provided.
        The query string is in the form of an SQL WHERE clause.

        :param selection_type: Indicates how results from the query should be handled
        by the selection set. Options are:
            NEW_SELECTION - The resulting selection replaces the current selection. This is the default.
            ADD_TO_SELECTION - The resulting selection is added to the current selection if one exists. If no selection
                exists, this is the same as the new selection option.
            REMOVE_FROM_SELECTION - The resulting selection is removed from the current selection. If no selection
                exists, this option has no effect.
            SUBSET_SELECTION - The resulting selection is combined with the current selection. Only records that are
                common to both remain selected.
            SWITCH_SELECTION - Switches the selection. All records that were selected are removed from the current
                selection; all records that were not selected are added to the current selection. The expression
                parameter is ignored when this option is specified.
            CLEAR_SELECTION - Clears or removes any selection. The expression parameter is ignored when this option is
                specified.
        :param query_string: An SQL expression used to select a subset of records.
        """
        # if an int is provided as a query string, assume it's to query the oid
        if isinstance(query_string, int):
            query_string = '"{}" = {}'.format(self.oid, query_string)

        # bust the currently cached query string
        self._cached_query_string = None
        self._cached_update_clause = None

        # clear the selections list if this is a new query
        if selection_type == 'NEW_SELECTION':
            self.clear_selection()

        # wrap field names in parens
        query_string = ' {}'.format(query_string)
        for field_name in self.field_names:
            query_string = re.sub(' {} '.format(field_name), ' "{}" '.format(field_name), query_string)
        query_string = query_string.strip()

        # add selection to list
        self._selections.append({
            'selection_type': selection_type,
            'query': query_string,
            'update_clause': query_string,
            'overlap_type': 'ATTRIBUTE',
            'from_tables': [f'"{self.schema}"."{self.table_name}"'],
            'search_objects': []
        })

    def get_selection_count(self):
        """ Gets the number of selected records in the table.

        If features are selected in the table, the function will return the number of selected. features. If no
        features are selected, the function will return the total number of features in the table.

        :return: The number of selected features in the table, or the total number of features in the table if there
        are no records selected.
        """
        # go ahead and retrieve the where clause (this forces
        # the execution of any spatial queries and insures
        # that our from clause is correct
        where_clause = self.where_clause
        query_string = 'SELECT count(distinct("{}")) as count FROM {} {}'.format(self.oid, self.from_clause, where_clause)

        query_result = self.workspace.execute_simple_sql(query_string)
        count = next(query_result)['count']

        return count

    def get_selection_set(self, field_names=None):
        """ Gets the selected features in the table in a SelectionSet object.

        :return: A `SelectionSet()` object containing the selected objects in the table.
        """
        field_names = self.field_names if field_names is None else field_names
        return SelectionSet(self, field_names, self.query_string())

    def switch_selection(self):
        """ Switches the selected records in the table.

        If records are selected in the table, this function will deselect the features that are currently selected,
        and will select the features that were previously unselected.
        """
        self.select_layer_by_attribute('', selection_type='SWITCH_SELECTION')

    def clear_selection(self):
        """ Clears the selected features in the table.

        Clears the table of any selected features.
        """
        # bust the currently cached query string
        self._cached_query_string = None
        self._cached_update_clause = None

        # reset the selection list
        self._selections = []

    def clone_selections(self):
        """ Returns a copy of the current selection set which can be reapplied to the layer using the
        `apply_selections()` method.

        :return: A list containing a copy of the current selections.
        """
        return self._selections[:]

    def apply_selections(self, selections):
        """ Applies the provided list of selections to the layer. Can be used with the `clone_selections()`
        method to back-up and reapply selections to the layer.

        :param selections: The list of selections to apply to the layer.
        :return: None.
        """
        # bust the currently cached query string
        self._cached_query_string = None
        self._cached_update_clause = None
        # copy the provided set of selections to the layer's container
        self._selections = selections[:]

    # SCHEMA MANAGEMENT METHODS
    def add_field(self, field, field_type='TEXT',
        default_value=None,
        field_length=50,
        geometry_type=None,
        spatial_reference=None,
        add_index=False):
        """ Adds a new field to the table.
        This method will add a new field to an existing table.
        :param field: The name of the new field or an ogr field object.
        :param field_type: The type of field. Valid values are:
            TEXT - Any string of characters.
            FLOAT - Fractional numbers between -3.4E38 and 1.2E38.
            DOUBLE - Fractional numbers between -2.2E308 and 1.8E308.
            SHORT - Whole numbers between -32,768 and 32,767.
            INTEGER - Whole numbers between -2,147,483,648 and 2,147,483,647.
            LONG - Whole numbers between -9.2E18 to 9.2E18.
            DATE - Date and/or time.
            BLOB - (not currently implemented) Long sequence of binary numbers. You need a custom loader or viewer
                or a third-party application to load items into a BLOB field or view the contents of a BLOB field.
            RASTER - (not currently implemented) Raster images. All ArcGIS software-supported raster dataset formats
                can be stored, but it is highly recommended that only small images be used.
            GUID - Globally unique identifier.
        :param default_value: The default value to set the field to after creating it. If not provided no value will
            be set.
        :param field_length: The field length when creating TEXT fields. This value has no effect when creating fields
            of other data types. If not specified when creating a TEXT field, the value will revert to the sde default
            (50 characters for most DBMS systems). (not currently implemented)
        """
        # if the field already exists simply return
        if field in self.field_names:
            return

        # translate the Postgresql data type
        if not field_type == 'GEOMETRY':
            if field_type == 'TEXT':
                field_length = 50 if field_length == 0 else field_length
                field_type = 'VARCHAR({})'.format(field_length)
            elif field_type == 'FLOAT' or field_type == 'DOUBLE':
                field_type = 'FLOAT8'
            elif field_type == 'SHORT':
                field_type = 'SMALLINT'
            elif field_type == 'INTEGER':
                field_type = 'INT'
            elif field_type == 'LONG':
                field_type = 'BIGINT'
            elif field_type == 'DATE':
                field_type = 'TIMESTAMP'
            elif field_type == 'GUID':
                field_type = 'UUID'
            alter_statement = f'ALTER TABLE "{self.schema}"."{self.table_name}" ADD COLUMN "{field}" {field_type}'
            self.workspace.execute_simple_sql(alter_statement, use_transaction=False)
        else:
            self.workspace.create_geometry_column(self.table_name, field, geometry_type, spatial_reference, schema=self.schema)
            self._cached_shape_field = None
            self._cached_geometry_columns = None
            self._cached_geometry_columns_reduced = None

        # reload the table
        self.alchemy_table = load_alchemy_table(self.workspace, self.table_name, schema=self.schema)

        if default_value is not None:
            self.calculate_field(field, default_value)

        if add_index and not field_type == 'GEOMETRY':
            self.add_index(field)

    def add_index(self, field_name):
        """ Adds an index to the specified attribute.

        :param field_name: A list of field names to be included in the index.
        :return: None
        """

        self.workspace.create_index(self.table_name, field_name, schema=self.schema)

    def list_indexes(self):
        """ List the names of the indexes associated with the table.

        :return: A list of index names.
        """
        return self.workspace.list_indexes(self.table_name, schema=self.schema)

    def delete_field(self, field_name):
        """ Deletes the specified field from the table.

        Deletes the specified field from the table. Any data contained in the deleted field will be lost.

        :param field_name: The field to be deleted.
        """
        field_names = [field_name] if isinstance(field_name, str) else field_name

        for field_name in field_names:
            shape_fields = self.get_geometry_field_names(include_hidden=True) if hasattr(self, 'get_geometry_field_names') else []
            assert field_name in self.field_names + shape_fields, 'Field does not exists'
            sql = f'ALTER TABLE "{self.schema}"."{self.table_name}" DROP COLUMN IF EXISTS "{field_name}"'
            self.workspace.execute_simple_sql(sql, use_transaction=False)

        # reload the table
        self.alchemy_table = load_alchemy_table(self.workspace, self.table_name, schema=self.schema)

    def delete_other_fields(self, preserve_fields):
        """
        Deletes all of the fields in the table except those specified in the `preserve_fields` set. Fields
        required by all shapefiles (FID, Shape, etc.) will be preserved automatically.

        :param preserve_fields: A set of fields not to be deleted.
        :return: List of fields that were deleted.
        """
        # if the field is a string make it a list
        if isinstance(preserve_fields, str):
            preserve_fields = [preserve_fields]

        #  add shapefile required fields to list of fields to preserve
        # preserve_fields_set = set(preserve_fields + constants.SHAPEFILE_NECESSARY_FIELDS + [self.oid])
        preserve_fields_set = set(preserve_fields + [self.oid])

        # get the list of fields in the table
        existing_fields = set(self.field_names)

        # get the list of fields to delete
        delete_fields = list(existing_fields - preserve_fields_set)

        # delete the fields in the table that are not in the preserve list
        for field_name in delete_fields:
            self.delete_field(field_name)
        return delete_fields

    def rename_field(self, field_name, new_field_name):
        """ Renames the selected field.

        :param field_name: The current name of the field to be renamed.
        :param new_field_name: The new name for the field.
        """
        existing_field_names = self.field_names

        assert field_name in existing_field_names, 'Could not find field {0}'.format(field_name)

        sql = f'ALTER TABLE "{self.schema}"."{self.table_name}" RENAME COLUMN "{field_name}" TO "{new_field_name}"'
        self.workspace.execute_simple_sql(sql, use_transaction=False)
        self.alchemy_table = load_alchemy_table(self.workspace, self.table_name, schema=self.schema)

    def trim_field(self, field):
        """ Trims the string values in the specified field.

        :param field: The name of the field to be trimmed.
        """
        assert field in self.field_names, 'Specified field must exists in the table.'
        self.calculate_field(field, 'TRIM ("{}")'.format(field))

    def delete(self):
        """ Deletes the underlying table.

        This method removes the table from ArcMap's tables list and deletes the underlying shapefile. Once delete has
        been called the table can no longer be accessed.

        """
        # do not process if table has already been deleted
        if self.deleted:
            return

        self.workspace.drop_table(self.table_name, schema=self.schema, cascade=True)
        self.deleted = True

    def get_field(self, field_name):
        """ Returns the field definition for the specified field name.

        :param field_name: The name of the field to be retrieved.
        :return: The field definition for the specified field.
        """

        for field in self.fields:
            if field.name == field_name:
                return field

    def list_fields(self, type='ALL'):
        """ Gets a list of fields in the table.

        This method will return a collection of field objects representing the fields that are in the current table.

        :return: An array of field objects representing the fields in the table.
        """
        return_list = []
        if type.upper() == 'ALL':
            for column in self.alchemy_table.columns:
                if column.name.lower() in ['edge_buffers', 'centroid', 'midpoint']:
                    continue
                return_list.append(column)
            return return_list

        for column in self.alchemy_table.columns:
            if (hasattr(self, 'get_geometry_field_names') and column.name in self.get_geometry_field_names(include_hidden=True)) or not hasattr(column.type, 'python_type'):
                continue
            if column.type.python_type == str and type.upper() == 'STRING':
                return_list.append(column)
            elif column.type.python_type == int and type.upper() == 'INTEGER':
                return_list.append(column)
            elif column.type.python_type == float and type.upper() == 'FLOAT':
                return_list.append(column)
        return return_list

    def list_field_names(self):
        """ Gets a list of field names in the table as strings. The list does *not* include the oid or shape
        field names (these can be accessed through the `Table.oid` and `Layer.shape` properties.

        :return: An array of strings representing the names of the fields in the table.
        """
        shape_fields = self.get_geometry_field_names(include_hidden=True) if hasattr(self, 'get_geometry_field_names') else []
        return [field.name for field in self.list_fields() if field.name != self.oid and field.name not in shape_fields]

    def get_oid_field(self):
        """ Gets the name of the OID field for this table or layer.

        The "OID" field is the primary key field in an Esri table or feature class. Typically,
        the name of the OID field in shapefiles is "FID" and the OID field in geodatabase tables
        and feature classes is called "OBJECTID".

        This method will dynamically determine the proper name of the field to be used in
        database queries, etc.

        :return: A string with the name of the OID field for this table or layer.
        """
        # if we've already determined the OID field name, just return it
        if self._cached_oid_field is not None:
            return self._cached_oid_field

        self._cached_oid_field = self.alchemy_table.primary_key.columns.items()[0][0]
        return self._cached_oid_field

    def get_unique_field_values(self, field,
        query_string=None):
        """ Gets a set of unique values found in the selected features' field.

        :param field: The field name to look up.
        :param query_string: An SQL expression used to select a subset of records.
        :return: A set of field values.
        """
        with self.get_search_cursor(field, query_string, sql_clause=('DISTINCT', None)) as cursor:
            return set([row[field] for row in cursor])

    def get_distinct_field_values(self, fields):
        """ Gets a set of unique values found in the selected features' field.

        :param fields: The field names to look up.
        :return: A set of field values.
        """
        if isinstance(fields, str):
            fields = [fields]
        field_string = ','.join(['"{}"'.format(field) for field in fields])
        sql = f'SELECT DISTINCT {field_string} FROM "{self.schema}"."{self.table_name}" {self.where_clause}'
        results = self.workspace.execute_simple_sql(sql)

        return_results = []
        for row in results:
            result = {}
            for field in fields:
                result[field] = row[field]
            return_results.append(result)
        return return_results

    def vacuum(self):
        """ Cleans up any data files produced by the table. Can be used to manage hard-drive space when
        doing large update operations.

        :return: None
        """

        self.workspace.vacuum(self.table_name, schema=self.schema)

    # CURSOR METHODS
    def get_search_cursor(self,
        field_names='*',
        query_string=None,
        sql_clause=(None, None),
        allow_invalid=False):
        """ Gets a fast read-only cursor for looping through features in the table.

        The cursor allows for row-by-row access to the features in the table. This cursor is optimized for reading and
        should not be used to update features. Use the get_update_cursor method to get a cursor that allows write-access
        to the features in the table.

        :param field_names: The field or list of fields to be returned in the cursor.
        :param query_string: An SQL expression used to select a subset of records.
        :param sql_clause: An optional pair of SQL prefix and postfix clauses organized in a list or tuple.

            An SQL prefix supports None, DISTINCT, and TOP. SQL postfix supports None, ORDER BY, and GROUP BY. An SQL
            prefix clause is positioned in the first position and will be inserted between the SELECT keyword and
            the SELECT COLUMN LIST. The SQL prefix clause is most commonly used for clauses such as DISTINCT or ALL.

            An SQL postfix clause is positioned in the second position and will be appended to the SELECT statement,
            following the where clause. The SQL postfix clause is most commonly used for clauses such as ORDER BY.

            The default value is (None, None)
        :return: A cursor that can be iterated over to review features in the table.
        """
        if query_string is None:
            query_string = self.query_string(allow_invalid=allow_invalid)
        return SearchCursor(self, field_names, query_string, sql_clause, allow_invalid)

    def get_insert_cursor(self, field_names='*'):
        """ Gets a fast cursor for adding features to the table.

        :param field_names: The field or list of fields to be returned in the cursor.
        :return: A cursor that can write new features to the table.
        """
        return InsertCursor(self, field_names)

    def get_update_cursor(self,
        field_names=None,
        query_string=None,
        sql_clause=(None, None),
        allow_invalid=False):
        """ Gets a cursor that allows for updating the features in the table.

        The update cursor allows the features in the cursor to be updated. If you simply need to loop through the
        features in the table but do not need to update them, use the get_search_cursor method.

        :param field_names: The field or list of fields to be returned in the cursor.
        :param query_string: An SQL expression used to select a subset of records.
        :param sql_clause: An optional pair of SQL prefix and postfix clauses organized in a list or tuple.

            An SQL prefix supports None, DISTINCT, and TOP. SQL postfix supports None, ORDER BY, and GROUP BY. An SQL
            prefix clause is positioned in the first position and will be inserted between the SELECT keyword and
            the SELECT COLUMN LIST. The SQL prefix clause is most commonly used for clauses such as DISTINCT or ALL.

            An SQL postfix clause is positioned in the second position and will be appended to the SELECT statement,
            following the where clause. The SQL postfix clause is most commonly used for clauses such as ORDER BY.

            The default value is (None, None)
        :return: A cursor that can be used to update features in the table.
        """
        if query_string is None:
            query_string = self.query_string(allow_invalid)
        return UpdateCursor(self, field_names, query_string, sql_clause, allow_invalid)

    # DATA MANIPULATION METHODS
    def calculate_field(self, field, expression):
        """ Populates a field in the table with calculated values.

        This method will calculate a value for each feature in the table and place that value in the specified field.
        The entry can be as simple as a static value that will be the same for all features, or it can be calculated
        individually based on values in each feature.

        :param field: The field where the output of the expression will be written.
        :param expression: The expression used to calculate the value.
        """
        query_string = f'UPDATE "{self.schema}"."{self.table_name}" SET "{field}" = {parse_expression(self, field, expression)} {self.update_clause}'

        self.workspace.execute_simple_sql(query_string)

    def calculate_fields(self, values):
        """ Updates multiple attributes for the selected set of records.

        :param values: A dictionary with keys representing the field names to update and values representing
            the updated values.
        :return: None
        """
        fields_string = ','.join(['"{}" = {}'.format(key, value) for key, value in values.items()])

        query_string = f'UPDATE "{self.schema}"."{self.table_name}" SET {fields_string} {self.update_clause}'

        self.workspace.execute_simple_sql(query_string)

    def copy_field_values(self, source_field, target_field):
        """ Copies the values of one field to another field.

        The copy_field_values method will copy the values from the field specified in the source_field parameter into
        the field specified in the target_field parameter. The value from the source field will be converted to a
        string, so the target field must be of type TEXT.
        :param source_field: The name of the field containing the source values.
        :param target_field: The name of the field that the values will be copied into. This field must be of type TEXT.
        """
        sql = f'UPDATE "{self.schema}"."{self.table_name}" SET "{target_field}" = "{source_field}" {self.where_clause};'
        self.workspace.execute_simple_sql(sql)

    def change_data_type(self, field, data_type, field_length=None):
        """ Changes the data type of the specified field to the specified data type, and cast the data within
        the field to that data type.

        :param field: The name of the field to be changed.
        :param data_type: The type of field to convert to. Valid values are:
            TEXT - Any string of characters.
            FLOAT - Fractional numbers between -3.4E38 and 1.2E38.
            DOUBLE - Fractional numbers between -2.2E308 and 1.8E308.
            SHORT - Whole numbers between -32,768 and 32,767.
            INTEGER - Whole numbers between -2,147,483,648 and 2,147,483,647.
            LONG - Whole numbers between -9.2E18 to 9.2E18.
            DATE - Date and/or time.
            BLOB - (not currently implemented) Long sequence of binary numbers. You need a custom loader or viewer
                or a third-party application to load items into a BLOB field or view the contents of a BLOB field.
            RASTER - (not currently implemented) Raster images. All ArcGIS software-supported raster dataset formats
                can be stored, but it is highly recommended that only small images be used.
            GUID - Globally unique identifier.
        :param field_length: If specified, text fields will be of the specified length. Default length is 50 if a length
            is not provided. This parameter is ignored for data types that do not require a length specification.
        """
        temp_field_name = '{}_old'.format(field)
        if data_type == 'TEXT':
            field_length = 50 if not field_length else field_length
            conversion_func = 'VARCHAR({})'.format(field_length)
        elif data_type == 'FLOAT' or data_type == 'DOUBLE':
            conversion_func = 'FLOAT8'
        elif data_type == 'SHORT':
            conversion_func = 'SMALLINT'
        elif data_type == 'INTEGER':
            conversion_func = 'INT'
        elif data_type == 'LONG':
            conversion_func = 'BIGINT'
        elif data_type == 'DATE':
            conversion_func = 'TIMESTAMP'
        elif data_type == 'GUID':
            conversion_func = 'UUID'
        else:
            conversion_func = data_type

        self.rename_field(field, temp_field_name)
        self.add_field(field, data_type,
                       field_length=field_length)
        self.calculate_field(field, 'CAST("{}" AS {})'.format(temp_field_name, conversion_func))
        self.delete_field(temp_field_name)

    def add_rows(self, rows):
        """ Add new rows to the table using an insert cursor.

        :param rows: A list of dictionaries where each dictionary represents a row to be added. The dictionary
            objects should have field names as keys with the corresponding data as dictionary values.
        :return: None
        """
        # if this is a single row, convert it to a list
        if isinstance(rows, dict):
            rows = [rows]

        # if an empty list is provided, just return
        if len(rows) == 0:
            return

        # get insertable rows for each provided row
        insertable_rows = [create_insert_row(self, row) for row in rows]

        with self.workspace.alchemy_engine.connect() as connection:
            connection.execute(self.alchemy_table.insert(), insertable_rows)

    def duplicate_table(self, output_table_name,
        output_workspace=None, output_schema=None):
        """
        Duplicates the layer's schema as a new empty shapefile. The data in the source shapefile is *not*
        copied to the new shapefile. Use the `copy_to_new_table` method if you wish to copy the source
        data into a new layer.

        :param output_table_name: The name of the new layer.
        :param output_workspace: The workspace where the output layer will be created.
            If not provided, the layer will be created in the same workspace as the existing layer.
        :return: The new empty layer
        """
        # set the workspace and schema if necessary
        output_workspace = self.workspace if output_workspace is None else output_workspace
        output_schema = self.workspace.schema if not output_schema else output_schema
        output_workspace.create_schema(output_schema)

        field_defs = []
        for field in self.fields:
            if field.name.lower() == self.oid.lower():
                field_def = Column(field.name, field.type, primary_key=True, autoincrement=True)
            else:
                field_def = Column(field.name, field.type)
            field_defs.append(field_def)

        alchemy_table = AlchemyTable(output_table_name, MetaData(), schema=output_schema, *field_defs)
        alchemy_table.create(output_workspace.alchemy_engine)

        return Table(output_table_name, output_workspace, schema=output_schema)

    def copy_to_new_table(self, output_table_name,
        output_workspace=None, output_schema=None):
        """ Copies selected rows to a new table.

        :param output_table_name: The name of the geodatabase table to write to.
        :param output_workspace: The workspace where the output layer will be created.
            If not provided, the layer will be created in the same workspace as the existing layer.
        :return: The Table object interface to the newly created table.
        """

        # set the workspace if necessary
        db_transfer = False
        if output_workspace is None:
            output_workspace = self.workspace
        else:
            db_transfer = self.workspace.connection_info['host'] != output_workspace.connection_info['host'] or \
                        self.workspace.connection_info['port'] != output_workspace.connection_info['port'] or \
                        self.workspace.connection_info['database'] != output_workspace.connection_info['database']

        # set the output schema if necessary
        output_schema = self.workspace.schema if not output_schema else output_schema

        self.duplicate_table(output_table_name, output_workspace, output_schema=output_schema)

        # go ahead and retrieve the where clause (this forces
        # the execution of any spatial queries and insures
        # that our from clause is correct
        where_clause = self.where_clause

        # copy fields
        if not db_transfer:
            query_string = 'INSERT INTO "{output_schema_name}"."{output_table_name}" ({output_field_names}) \
                SELECT {field_names} FROM {from_clause} {where_clause}'\
                .format(
                output_schema_name=output_schema,
                output_table_name=output_table_name,
                output_field_names=','.join(['"{}"'.format(field_name) for field_name in self.field_names]),
                field_names=','.join(['"{}"'.format(field_name) for field_name in self.field_names]),
                from_clause=self.from_clause,
                where_clause=where_clause)
        else:
            # because where clause in this case is in a string it must be formatted
            where_clause = where_clause.replace("'", "''")
            query_string = "INSERT INTO \"{output_schema_name}\".\"{output_table_name}\" ({output_field_names}) \
                SELECT {field_names} FROM dblink('host={host} port={port} dbname={database} \
                user={user} password={password} options=-csearch_path=', 'select {field_names} from {from_clause} {where_clause}') \
                as link_table({field_definitions})"\
                .format(
                output_schema_name=output_schema,
                output_table_name=output_table_name,
                output_field_names=','.join(['"{}"'.format(field_name) for field_name in self.field_names]),
                field_names=','.join(['"{}"'.format(field_name) for field_name in self.field_names]),
                from_clause=self.from_clause,
                where_clause=where_clause,
                host=self.workspace.connection_info['host'],
                database=self.workspace.connection_info['database'],
                port=self.workspace.connection_info['port'],
                user=self.workspace.connection_info['user'],
                password=self.workspace.connection_info['password'],
                field_definitions=','.join(['"{}" {}'.format(field_name, self.get_field_type(field_name)) for field_name in self.field_names]))

        output_workspace.execute_simple_sql(query_string)

        return Table(output_table_name, output_workspace, schema=output_schema)

    def truncate(self):
        """
        Deletes the selected records in the table.
        :return: None
        """
        query_string = 'TRUNCATE TABLE  "{}"."{}"'.format(self.schema, self.table_name)

        self.workspace.execute_simple_sql(query_string)
        self.clear_selection()

    def delete_selected_rows(self):
        """
        Truncates the table.
        :return: None
        """
        query_string = 'DELETE FROM "{}"."{}" {}'.format(self.schema, self.table_name, self.where_clause)

        self.workspace.execute_simple_sql(query_string)
        self.clear_selection()

    def delete_duplicates(self, fields):
        """ Deletes duplicate records in the table.

        This method deletes duplicate records in the table based on the list of fields provided. If the field Shape is
        selected, feature geometries are compared.

        :param fields: The fields to be compared.
        """
        query_string = 'DELETE FROM "{schema_name}"."{table_name}" WHERE "{oid}" NOT IN \
                        (SELECT min("{oid}") AS OID FROM {table_name} GROUP BY {group_fields})'\
                        .format(
            oid=self.oid,
            schema_name=self.schema,
            table_name=self.table_name,
            group_fields=','.join(['"{}"'.format(field) for field in fields]),
        )

        self.workspace.execute_simple_sql(query_string)

    def _assert_matching_fields(self, other_table):
        """ Insures that the field names between this table and the other table match

        :param other_table: The table to be compared to this table.
        :return: None. Assertion errors will raise if the fields do not match.
        """
        source_field_names = [field_name.lower() for field_name in self.field_names]
        input_field_names = [field_name.lower() for field_name in other_table.field_names]
        for field_name in input_field_names:
            assert field_name in source_field_names

    def append(self, add_tables, schema_type='TEST'):
        """ Appends features from one or more tables to this table. This mutates the table.

        :param add_tables: A list of tables with features to append to this table.
        :param schema_type: A string specifying if the schema of the tables should match. Valid options are:
            TEST - Schema must match, otherwise throw error.
            NO_TEST - Schema do not have to match.
        """
        if not isinstance(add_tables, list):
            add_tables = [add_tables]
        if schema_type == 'TEST':
            for table in add_tables:
                self._assert_matching_fields(table)
        shape_field = self.shape if hasattr(self, 'shape') else None
        fields_to_copy = [name for name in self.field_names if name != self.oid]
        if shape_field is not None:
            fields_to_copy.append(shape_field)
        with self.get_insert_cursor(fields_to_copy) as insert_cursor:
            for add_table in add_tables:
                with add_table.get_search_cursor(fields_to_copy) as search_cursor:
                    for row in search_cursor:
                        insert_cursor.insert_row(row)

    def to_csv(self, path, attributes):
        """ Writes table contents to csv file.

        :param path: Path and filename of csv file to write to.
        :param attributes: The attributes (columns) to include in the csv.
        """
        with open(path, 'w') as output:
            output.write(','.join(attributes) + '\n')
            with self.get_search_cursor(attributes) as cursor:
                output.write('\n'.join(
                    [','.join([str(row[field]) for field in row])
                    for row in cursor]))

    def get_field_type(self, field_name):
        """ Gets the data type for the specified field.

        :param field_name: The name of the field to inspect.
        :return: The data type for the field. Valid values are:
            Blob - Blob
            Date - Date
            Double - Double
            Geometry - Geometry
            GlobalID - Global ID
            Guid - Guid
            Integer - Integer (Long Integer)
            OID - Object ID
            Raster - Raster
            Single - Single (Float)
            SmallInteger - Small Integer (Short Integer)
            String - String (Text)
        """
        shape_fields = self.get_geometry_field_names(include_hidden=True) if hasattr(self,
                                                                                     'get_geometry_field_names') else []
        field_names = self.field_names + shape_fields
        if field_name not in field_names:
            raise ValueError('Field not found - {}'.format(field_name))
        field_def = self.get_field(field_name)

        field_type = str(field_def.type)
        return field_type.replace('_', ' ')

    @property
    def max_oid(self):
        """ Returns the highest value in the designated `oid` field.

        :return: An integer representing the maximum value of the `oid` field.
        """
        result = self.workspace.execute_simple_sql('SELECT MAX("{}") as max_oid FROM "{}"."{}"'.format(self.oid, self.schema, self.table_name))
        row = next(result)
        return int(row['max_oid']) if row is not None and row['max_oid'] is not None else 0

    def query_string(self, allow_invalid=False):
        """ The query string for the selection set applied to the table.

        :return: A string representing the WHERE clause of the selection set applied to the table.
        """
        # return cached string if available
        if self._cached_query_string:
            return self._cached_query_string

        # cache new query string and return it
        self._cached_query_string = build_query_string(self, self._selections, allow_invalid=allow_invalid)
        return self._cached_query_string

    @property
    def update_clause(self):
        """ Returns the WHERE clause required to update the data in the table.

        :return: A string representing the WHERE clause required to update the table.
        """
        if self._cached_update_clause:
            return self._cached_update_clause

        # make sure the main query string has been cached (required for the update clause to work correctly)
        self._cached_query_string = build_query_string(self, self._selections)

        self._cached_update_clause = build_update_clause(self, self._selections)
        return self._cached_update_clause

    @property
    def from_clause(self):
        """ Properly formats FROM clause for SQL queries using the layer's selection filters.

        :return: String.
        """
        if len(self._selections) == 0:
            return f'"{self.schema}"."{self.table_name}"'

        last_selection = self._selections[-1]
        if 'result_query_string' in last_selection:
            return ", ".join(last_selection['from_tables'])

        from_tables = set()
        for selection in self._selections:
            from_tables.update(selection['from_tables'])
        return ", ".join(from_tables)

    @property
    def where_clause(self):
        """ Properly formats WHERE clause for SQL queries using the layer's selection filters.

        :return: String, a clause to append to end of SQL query.
        """
        return "WHERE {}".format(self.query_string(allow_invalid=True)) if len(self._selections) > 0 else ""

    def get_field_length(self, field_name):
        """ Gets the field length for the specified field.

        :param field_name: The name of the field to inspect.
        :return: The length of the field.
        """
        sql = "SELECT character_maximum_length FROM information_schema.COLUMNS WHERE table_schema = '{}' AND table_name = '{}' AND column_name = '{}';".format(self.schema, self.table_name, field_name)

        result = self.workspace.execute_simple_sql(sql)
        row = next(result)
        return int(row['character_maximum_length']) if row is not None and row['character_maximum_length'] is not None else 0

    # properties and support methods
    count = property(fget=get_selection_count)
    fields = property(fget=list_fields)
    oid = property(fget=get_oid_field)
    field_names = property(fget=list_field_names)


class Table(TableBase):
    """
    The table class manages table names and the association
    with its underlying shapefile. It also provides access
    to most geoprocessing tools that take tables as inputs.
    """

    def __init__(self, table, workspace, schema=None):
        """ Initializes the table object.

        All "Table" functions are actually housed in the `TableBase` class. The `Table` class is a helper class
        that allows a Table object to be initialized correctly in ArcPy.

        :param table_name: The name of the table as it will be displayed in the ArcMap table control (this is the
            internal name of the table)
        :param workspace: The workspace where the table is located.
        :param file_path: The path to the underlying dbf that stores the data for this table, if not in a geodatabase.
        """
        super(Table, self).__init__(table, workspace, schema=schema)

    @staticmethod
    def merge_tables_to_table(tables, output_table_name, output_workspace):
        """ Merges the specified tables into a new table.

        :param tables: The table objects to be merged.
        :param output_table_name: The name of the output table that will contain the processed data.
        :param output_workspace: The workspace where the output layer will be located.
        :return: A new table object that represents the table containing the results of the process.
        """
        first_table = tables.pop()
        first_table.clear_selection()
        return_table = first_table.copy_to_new_table(output_table_name, output_workspace)

        for next_table in tables:
            return_table.append(next_table)

        return return_table



# coding=utf-8
""" Contains objects related to lines.

"""
import itertools
# import logging
import math
import numpy as np

from itertools import islice
from more_itertools import pairwise
from osgeo.ogr import CreateGeometryFromWkt as wkt_to_ogr
from shapely.coords import CoordinateSequence
from shapely.ops import linemerge
from shapely.geometry import Point as ShapelyPoint, LineString, MultiLineString
from shapely.wkt import loads as wkt_to_geom
from src.oogeo.objects.multipointbase import MultipointBase
from src.oogeo.functions.conversions import upgrade_shapely_geometry
import src.oogeo.objects.point
import src.oogeo.objects.polygon


def correct_invalid_line(shapely_line, spatial_reference=2263):
    """

    :param shapely_polygon:
    :return:
    """

    # break the line down to edges
    duplicate_points = []
    lines = []
    for line in shapely_line.geoms:
        line_edges = []
        for point_1, point_2 in pairwise(list(line.coords)):
            point_1 = src.oogeo.objects.point.Point(point_1, spatial_reference)
            point_2 = src.oogeo.objects.point.Point(point_2, spatial_reference)
            if point_1.equals(point_2):
                duplicate_points.append(point_2)
                continue
            line_edges.append(src.oogeo.objects.line.Line([point_1, point_2], spatial_reference))
        lines.append(line_edges)

    # compare edges and find removal points
    removal_points = []
    for line_edges in lines:
        for edge_1, edge_2 in pairwise(line_edges):
            angle_1 = edge_1.endpoints_angle + 360 if edge_1.endpoints_angle < 0 else edge_1.endpoints_angle
            angle_2 = edge_2.endpoints_angle + 360 if edge_2.endpoints_angle < 0 else edge_2.endpoints_angle
            angle_diff = abs(angle_1 - angle_2) - 180
            if -1 < angle_diff < 1:
                removal_points.append(edge_2.first_point)

    # reconstruct line and remove offensive points
    new_lines = []
    for line in shapely_line.geoms:
        line_points = []
        for point in line.coords:
            point = src.oogeo.objects.point.Point(point, spatial_reference)
            remove = False
            for removal_point in removal_points:
                if point.equals(removal_point):
                    remove = True
                    break
            if not remove:
                for duplicate_point in duplicate_points:
                    if point.equals(duplicate_point):
                        duplicate_points.remove(duplicate_point)
                        remove = True
                        break
            if not remove:
                line_points.append(point.geometry)
        new_lines.append(LineString(line_points))

    if len(new_lines) == 0:
        return None
    return MultiLineString(new_lines)


def has_consecutive_near_points(test_line, other_line):
    """ Determines if two or more consecutive vertices of the test line are near the other line.
    This indicates that the two lines share a line segment.

    :param test_line: The line to be evaluated.
    :param other_line: The line for comparison.
    :return: True if two or more consecutive vertices of the test line are near the other line, otherwise False.
    """
    first_near_found = False
    for point in test_line.points:
        # get the distance from this vertex to the other edge -
        # if we're < 0.001 feet away, consider the vertex to be
        # on the edge
        test_distance = point.distance_to(other_line)
        if test_distance < 0.001:
            if not first_near_found:
                # flag that we found a the first near vertex
                first_near_found = True
            else:
                # if we're here it means we found two near vertices
                # in a row, so share a line segment with is true
                return True
        else:
            first_near_found = False
    return False


def unloop(linestring):
    """ Removes either starting or ending segment if the linestring is detected to be a loop.

    :param linestring: Shapely LineString object
    :return: Shapely LineString object
    """
    v = linestring.coords
    if v[0] != v[-1]:
        # Not a loop
        return linestring

    # logging.warning('Warning - loop line detected - correcting.')

    # find the longest line segment
    start_length = (v[1][0] - v[0][0]) ** 2 + (v[1][1] - v[0][1]) ** 2
    end_length = (v[-1][0] - v[-2][0]) ** 2 + (v[-1][1] - v[-2][1]) ** 2

    # remove a segment based on the lengths
    # logging.debug('start_length^2 = %s, end_length^2 = %s', start_length, end_length)
    if start_length > end_length:
        # logging.debug('Removing start segment')
        return LineString(v[1:])
    else:
        # logging.debug('Removing end segment')
        return LineString(v[:-1])


class Line(MultipointBase):
    """ Provides support for analyzing polyline features in shapefiles.

    The Line class abstracts some of the logic in accessing attribute of polyline features from shapefiles,
    while providing support methods for common operations.
    """

    def __init__(self, geometry, spatial_reference, allow_invalid=False):
        """ Initializes the line object.

        :param geometry: The geometry object representing a polyline. May be a list of coordinates (as tuples
            or dicts), ogr LineString, or Shapely LineString.
        :param spatial_reference: The spatial reference of the line.
        :param allow_invalid: If true, errors for invalid geometries will be ignored. Set this
            flag to true when loading datasets where you expect to encounter invalid geometries.
        """

        if isinstance(geometry, str):
            geometry = wkt_to_geom(geometry)
        if isinstance(geometry, CoordinateSequence):
            geometry = LineString(geometry)
        if isinstance(geometry, list):
            if not isinstance(geometry[0], list):
                points = []
                for point in geometry:
                    if isinstance(point, tuple):
                        points.append((point[0], point[1]))
                    elif isinstance(point, dict):
                        points.append((point['x'], point['y']))
                    elif isinstance(point, src.oogeo.objects.point.Point):
                        points.append((point.x, point.y))
                geometry = LineString(points)
            else:
                parts = []
                for part in geometry:
                    points = []
                    for point in part:
                        if isinstance(point, tuple):
                            points.append((point[0], point[1]))
                        elif isinstance(point, dict):
                            points.append((point['x'], point['y']))
                        elif isinstance(point, src.oogeo.objects.point.Point):
                            points.append((point.x, point.y))
                    parts.append(LineString(points))
                geometry = MultiLineString(parts)

        assert isinstance(geometry, LineString) or isinstance(geometry, MultiLineString), 'A line could not be created.'

        # elevate the geometry to a multigeometry if necessary
        if isinstance(geometry, MultiLineString):
            geometry = linemerge(geometry)
        if isinstance(geometry, LineString):
            geometry = MultiLineString([geometry])

        super(Line, self).__init__(geometry, spatial_reference, allow_invalid)

        self._segments = None
        self._cached_chord = None
        self._cached_points = None

    @staticmethod
    def create_ray(start_point, angle, length):
        """ Creates a `Line()` object beginning at the specified point and extending at the specified angle for
        the specified length.

        Note: This formula was found on GitHub here: https://gist.github.com/karolba/71755c3ac66c58ee02c095bd70a82457
        **It does not support Geographic projections**

        :param start_point: The start point to begin the line.
        :param angle: The angle at which to draw the line.
        :param length: The length of the line.
        :return: A `Line()` object beginning at the specified point and extending at the specified angle for
        the specified length.
        """
        x = start_point.x + math.sin(math.radians(angle)) * length
        y = start_point.y + math.cos(math.radians(angle)) * length
        end_point = src.oogeo.objects.point.Point((x, y), start_point.spatial_reference)
        return Line([start_point, end_point], start_point.spatial_reference)

    @staticmethod
    def count_right_angles(lines, tolerance=5):
        """ Counts the number of right angles between the lines in the provided list.

        :param lines: List of `Line` objects to compare.
        :param tolerance: Number of degrees that a right angle can deviate from 90 or -90.
        :return: Number of right angles between the lines.
        """
        if len(lines) < 2:
            return 0

        # iterate through the lines and check the angles of the ones that touch
        angles = (a.angle_between(b)
            for a, b in itertools.combinations(lines, 2) if a.touches(b))

        return sum(abs(angle - 90) <= tolerance for angle in angles)

    @staticmethod
    def is_parallelogram(lines):
        """ Analyzes the provided list of lines and determines if they can form a parallelogram.

        :param lines: The list of `Line` objects to be evaluated.
        :return: True if the lines form a parallogram, otherwise False.
        """
        return len(lines) == 4 and Line.count_parallel_sets(lines) == 2

    @staticmethod
    def count_parallel_sets(lines):
        """ Counts the number of parallel, non-adjacent lines in a list of lines.

        :param lines: The list of `Line` objects to be evaluated.
        :return: The number of parallel lines found in the list of lines.
        """
        parallel_line_count = 0
        for line_1, line_2 in itertools.combinations(lines, 2):
            # if the lines touch do not compare them
            if line_1.touches(line_2):
                continue
            if line_1.is_parallel(line_2, tolerance=10):
                parallel_line_count += 1

        return parallel_line_count

    @staticmethod
    def angle(p1, p2):
        """ Returns the angle between two points. Assumes a planar measurement method.

        Based on previous conventions, this returns the angle of the vector formed from point one
        to point two, measured clockwise from the positive y-axis.

        :param p1: The first point, as a `Point()` object.
        :param p2: The second point, as a `Point()` object.
        :return: Number, ranging 0-360.
        """
        dx = p2.x - p1.x
        dy = p2.y - p1.y
        if dx == 0 and dy == 0:
            # logging.warning('Points are coincident, returning zero for angle.')
            return 0
        angle = 90 - np.rad2deg(np.arctan2(dy, dx))
        return angle if angle >= 0 else 360 + angle

    @staticmethod
    def degrees_off_parallel(degrees):
        """ Returns absolute degrees removed from 0 or 180, whichever is closer.

        :param degrees: Number of degrees.
        :return: Number, ranging 0 to 90.
        """

        if np.isnan(degrees):
            # this happens when we get an invalid geometry.
            # log it and return zero for now (need to think
            # about a better way to handle it).
            # logging.warning('Invalid input - cannot determine degrees of parallel for NaN')
            return 0
        if 0 <= degrees <= 90:
            return degrees
        elif 90 < degrees <= 180:
            return 180 - degrees
        elif 180 < degrees <= 360:
            return Line.degrees_off_parallel(degrees - 180)
        elif degrees > 360:
            return Line.degrees_off_parallel(degrees % 360)
        else:
            return Line.degrees_off_parallel(-degrees)

    @property
    def chord(self):
        """

        :return:
        """
        if self._cached_chord is None:
            self._cached_chord = np.array((self.last_point.x, self.last_point.y)) - np.array(
                (self.first_point.x, self.first_point.y))
        return self._cached_chord

    @property
    def first_point(self):
        """ Returns coordinates for the first vertex of this Line.

        :return: A `Point()` object representing the first point in the line.
        """
        # get the first point from the first geometry
        return self.get_point(0)

    @property
    def last_point(self):
        """ Returns coordinates for the last vertex of this Line.

        :return: A `Point()` object representing the last point in the line.
        """
        return self.get_point(self.point_count - 1)

    @property
    def chord_length(self):
        """ The straight-distance length between first and last points of the Line.

        :return: Number.
        """
        return np.linalg.norm(self.chord)

    @property
    def first_segment(self):
        """ The first segment of a multi-segmented polyline (or the entire line if the line on
        has a single segment).

        :return: A `Line()` object representing the first segment of a multi-segment polyline.
        """
        # get the first segment from the first geometry
        first_point = self.get_point(0)
        second_point = self.get_point(1)
        return Line([first_point, second_point], self.spatial_reference)

    @property
    def last_segment(self):
        """ The last segment of a multi-segmented polyline (or the entire line if the line on
        has a single segment).

        :return: A `Line()` object representing the last segment of a multi-segment polyline.
        """
        # get the last segment from the last geometry
        first_point = self.get_point(self.point_count - 2)
        second_point = self.get_point(self.point_count - 1)
        return Line([first_point, second_point], self.spatial_reference)

    @property
    def length(self):
        """ Returns the length of this Line, i.e. the sum of all segment lengths.

        :return: Number.
        """
        total = 0
        for linestring in self.geometry.geoms:
            total += linestring.length
        return total

    @property
    def is_closed(self):
        """ Indicates if the line is closed (begins and ends at the same vertex).

        :return: True of the line is closed, otherwise False.
        """
        return True if self.first_point.equals(self.last_point) else False

    @property
    def endpoints_angle(self):
        """ Returns the angle in degrees from this line's first vertex to last vertex.

        :return: Number.
        """
        return Line.angle(self.first_point, self.last_point)

    @property
    def endpoints_theta(self):
        """ Returns the theta angle in degrees from this line's first vertex to last vertex.

        :return: Number.
        """
        delta_x = self.last_point.x - self.first_point.x
        delta_y = self.last_point.y - self.first_point.y
        theta_radians = math.atan2(delta_y, delta_x)
        return math.degrees(theta_radians)

    @property
    def first_angle(self):
        """ Returns the angle in degrees from this line's first vertex to second vertex.

        :return: Number.
        """
        return Line.angle(self.first_point, self.get_point(1))

    @property
    def last_angle(self):
        """ Returns the angle in degrees from this line's second-to-last vertex to last vertex.

        :return: Number.
        """
        return Line.angle(self.get_point(self.point_count - 2), self.last_point)

    @property
    def centroid(self):
        """ Returns coordinates for point representing the centroid of the line.

        :return: A `Point()` object representing the centroid of the geometry.
        """
        return src.oogeo.objects.point.Point(self.geometry.centroid, self.spatial_reference)

    @property
    def turn_count(self):
        """ Returns the number of line segments that turn the overall line substantially (>45 degrees).

        :return: Number of turns greater than 45 degrees in the line.
        """
        # get the line segments for the line
        angles = (a.angle_between(b)
            for a, b in self.get_line_segment_pairs())

        return sum(angle > 45 for angle in angles)

    def _get_cached_points(self):
        """ Caches the vertices of the geometry as points and insures that the points are only collected once.

        :return: A list of `Point()` objects representing the vertices of the geometry.
        """
        if self._cached_points is not None:
            return self._cached_points

        self._cached_points = []
        for linestring in self.geometry.geoms:
            for point in linestring.coords:
                self._cached_points.append(src.oogeo.objects.point.Point(point, self.spatial_reference))

        return self._cached_points

    def get_point(self, index):
        """ Returns coordinates for vertex at the provided index of this Line.

        :param index: The integer index of the desired vertex. Negative integers will count backward from the last
            vertex, with -1 returning the last point.
        :return: Tuple (length 2) with x-y coordinates.
        """
        return self.points[index]

    @property
    def point_count(self):
        """ Returns the number of points in the geometry.

        :return: The number of points in the geometry.
        """
        return len(self.points)

    @property
    def points(self):
        """ Returns a list of coordinates representing the vertices of the line.

        :return: A list of `Point()` objects representing the points in the line.
        """
        return self._get_cached_points()

    def split_line_at_turns(self, tolerance=45):
        """ Splits the line into line segments based on where the line turns.

        :param tolerance: The angle which defines a "turn" in degrees. The default is 45 degrees.
        :return: A list of `Line()` objects representing the segments of the split line.
        """

        # initialize the return array
        return_list = []

        # get the line segments for the line
        line_segments = self.split_line_at_vertices()

        if len(line_segments) == 1:
             return line_segments

        segment_a = line_segments.pop(0)
        for segment_b in line_segments:
            if not segment_a.touches(segment_b):
                return_list.append(segment_a)
                segment_a = segment_b
                continue
            turn_angle = segment_a.angle_between(segment_b)
            if tolerance <= turn_angle < (360 - tolerance):
                return_list.append(segment_a)
                segment_a = segment_b
                continue
            segment_a = segment_a.union(segment_b)
        return_list.append(segment_a)

        # at this point, segment a is now the last segment in the list -
        # compare it to the first to make sure we went all the way around the edges
        if len(return_list) > 2:
            first_segment = return_list[0]
            last_segment = return_list[-1]
            if first_segment.touches(last_segment):
                turn_angle = first_segment.angle_between(last_segment)
                if tolerance <= turn_angle < (360 - tolerance):
                    return_list[0] = first_segment.union(last_segment)
                    return_list = return_list[:-1]

        return return_list

    def turn_direction(self, next_line):
        """ Indicates if the "next" line turns to the left or right relative to the current line.

        :param next_line: The line to be compared to the current line.
        :return: -1 if the next line turns to the left relative to the current line or +1 if it turns to the right
            (or 0 if exactly straight, which is very rare).
        """
        # (turn flag is -1 if turning to the left, +1 if turning to the right)

        direction = -1 * np.cross(self.chord, next_line.chord)
        return 1 if direction > 0 else -1 if direction < 0 else 0

    @property
    def is_curve(self):
        """ Indicates if segments in the polyline represent a curve.

        :return: True if the segments in the polyline have a consistent curve, else False.
        """
        # constants that may need tweaking
        min_curve_count = 4
        max_length = 40

        # if there are less than 4 segments then the line cannot curve
        if self.point_count < min_curve_count + 1:
            return False

        # get the line segments for the line
        segment_pairs = self.get_line_segment_pairs()

        turns = ({
            'within_length': segment_a.length < max_length and segment_b.length < max_length,
            'direction': segment_a.turn_direction(segment_b),
            'angle': segment_a.angle_between(segment_b)
        } for segment_a, segment_b in segment_pairs)

        last_turn_direction = 0
        curve_count = 0
        for turn in turns:
            if (turn['within_length']
                and 0.8 <= turn['angle'] <= 10
                and turn['direction'] != 0
                and turn['direction'] + last_turn_direction != 0):
                curve_count += 1
                if curve_count >= min_curve_count:
                    return True
                last_turn_direction = turn['direction']
            else:
                curve_count = 0
                last_turn_direction = 0

        return False

    @property
    def midpoint(self):
        """ Returns the midpoint of the line.

        :return: The midpoint of the line.
        """
        return self.point_at_distance(0.5, True)

    def is_parallel(self, other_line, tolerance=45):
        """ Determine whether line is parallel with another line, within some angle tolerance.

        This method measures the angle between the line segments formed by the respective endpoints of this Line object
        and another provided Line. It then determines if the line segments are parallel within some tolerance, measured
        in degrees.

        :param other_line: A Line object to compare angles with.
        :param tolerance: The tolerance in degrees away from 0 or 180 degree angle formed by the line segments.
        :return: Boolean.
        """
        a = Line.degrees_off_parallel(self.angle_between(other_line))
        return a <= tolerance

    def is_perpendicular(self, other_line, tolerance=45):
        """ Determine whether line is parallel with another line, within some angle tolerance.

        This method measures the angle between the line segments formed by the respective endpoints of this Line object
        and another provided Line. It then determines if the line segments are parallel within some tolerance, measured
        in degrees.

        :param other_line: A Line object to compare angles with.
        :param tolerance: The tolerance in degrees away from 0 or 180 degree angle formed by the line segments.
        :return: Boolean.
        """
        tolerance = 90 - tolerance
        a = Line.degrees_off_parallel(self.angle_between(other_line))
        return a >= tolerance

    def generalize(self, max_offset=1):
        """ Reduces the number of vertices in the line.

        Generalizing a line alters its path and will cause it to deviate from the original line depending on
        how many vertices are removed. The `max_offset` parameter specifies a maximum distance (in map units)
        that the line will be permitted to deviate from the original path. As the `max_offset` value grows larger,
        there will be fewer vertices in the resulting line, and the resulting line will deviate more from the
        original. Under the hood, this uses Shapely's `simplify` method.

        :param max_offset: The maximum distance that the line will be permitted to deviate
            from the original path.
        :return: A `Line` object representing the simplified line.
        """
        self._segments = None
        return Line(self.geometry.simplify(max_offset), self.spatial_reference)

    def densify(self, distance):
        """ Add vertices to the line using the specified method.

        :param distance: The maximum distance between vertices. The actual distance between vertices will usually
            be less than the maximum distance as new vertices will be evenly distributed along the original segment.
            If using a type of DISTANCE or ANGLE, the distance is measured in the units of the geometry's spatial
            reference. If using a type of GEODESIC, the distance is measured in meters.
        :return: A new `Line` object representing the densified line.
        """
        ogr_geometry = wkt_to_ogr(self.wkt)
        ogr_geometry.Segmentize(distance)
        return Line(wkt_to_geom(ogr_geometry.ExportToWkt()), self.spatial_reference)

    def intersect(self, other, return_geometry_type, exact=False):
        """ Returns a geometry representing the intersection of this geometry and the provided geometry. The
        method will return the specified geometry type ('POINT', 'POLYLINE', or 'POLYGON')

        :param other: The geometry to be compared to this geometry.
        :param return_geometry_type: The type of geometry to be returned. Valid values are:
            * 'POINT': A point representing the intersection point of the two geometries will be returned.
            * 'POLYLINE': A polyline will be returned, representing the edges where the two geometries intersect.
            * 'POLYGON': A polygon will be returned, representing the area where the two geometries intersect.
        :param exact: If True, the geometries will be compared exactly as they are. If False, the vertices
            of the geometries will be merged prior to comparison to mitigate for tiny gaps and overlaps.
        :return: A geometry object of the specified type representing the intersection of the two geometries.
        """

        if not exact:
            merged_self, merged_other = MultipointBase.merge_vertices(self, other)
        else:
            merged_self = self
            merged_other = other

        # return None if the geometries do not intersect at all
        if merged_self.disjoint(merged_other):
            return None

        intersection_geometries = merged_self.geometry.intersection(merged_other.geometry)
        return_geometries = []
        if not isinstance(intersection_geometries, list):
            intersection_geometries = [intersection_geometries]
        for geometry in intersection_geometries:
            if isinstance(geometry, ShapelyPoint):
                if not return_geometry_type == 'POINT':
                    continue
                return_geometries.append(src.oogeo.objects.point.Point(geometry, self.spatial_reference))
            elif isinstance(geometry, LineString) or isinstance(geometry, MultiLineString):
                line = Line(geometry, self.spatial_reference)
                if return_geometry_type == 'POLYLINE':
                    return_geometries.append(line)
                else:
                    return_geometries.append(line.first_point)

        if len(return_geometries) == 0:
            return None
        if len(return_geometries) == 1:
            return return_geometries[0]
        return return_geometries

    def snap_point(self, point):
        """ Returns the point along the line that is nearest to the indicated point.

        :param point: The point to be snapped.
        :return: `Point()` object representing a point along the line that is nearest to the specified point.
        """
        distance_along_line = self.geometry.project(point.geometry)
        return self.point_at_distance(distance_along_line)

    def point_at_distance(self, distance, use_percentage=False):
        """ Returns a PointGeometry object representing the point along the line that is the specified distance
        from the beginning of the line. The distance can be expressed in both map units and as a percentage of
        the length of the line if the `use_percentage` flag is set.

        If the `distance` value is less than zero the start point of the line will be returned. If the `distance`
        value is greater than the length of the line the end point of the line will be returned.

        :param distance: The distance to measure from the beginning of the line. The distance can be expressed in
        both map units and as a percentage of the length of the line if the `use_percentage` flag is set. The
        percentage should be expressed as a decimal value ranging from 0 to 1.
        :param use_percentage: If `True` the `distance` parameter will be interpreted as a percentage along the
        line. The percentage should be expressed as a decimal value ranging from 0 to 1.
        :return: A `Point()` object representing the point at the specified distance along the line.
        """
        return src.oogeo.objects.point.Point(self.geometry.interpolate(distance, use_percentage), self.spatial_reference)

    def add_point(self, point, allow_invalid=False):
        """ Adds the specified point to the geometry.

        :param point: The point to be added.
        :return: A new geometry object with the new point inserted into the points list.
        """

        # look through the line to see if we already have this point

        # find the nearest points to the specified point
        nearest_index = None
        for edge_index, edge in enumerate(self.geometry.geoms):
            for point_index in range(len(edge.coords)):
                # if this is the last vertex, associate it with the first one to make an edge
                if point_index == len(edge.coords) - 1:
                    vertex_pair = [src.oogeo.objects.point.Point(edge.coords[point_index], self.spatial_reference),
                                   src.oogeo.objects.point.Point(edge.coords[0], self.spatial_reference)]
                else:
                    vertex_pair = [src.oogeo.objects.point.Point(coords, self.spatial_reference) for coords in
                                   islice(edge.coords, point_index, point_index + 2)]

                # if this point equals the test point, do not add the point
                # simply return a copy of the main geometry
                if vertex_pair[0].equals(point) or vertex_pair[0].distance_to(point) < 0.000001:
                    return Line(MultiLineString(self.geometry), self.spatial_reference)

                if vertex_pair[0].equals(vertex_pair[1]):
                    continue
                test_line = Line(vertex_pair, self.spatial_reference)

                test_index = {
                    'edge_index': edge_index,
                    'point_index': point_index,
                    'distance': point.distance_to(test_line)
                }
                if not nearest_index or test_index['distance'] < nearest_index['distance']:
                    nearest_index = test_index

        # create the new geometry
        shapely_lines = []
        for edge_index, edge in enumerate(self.geometry.geoms):
            new_edge_points = []
            for point_index, test_point in enumerate(edge.coords):
                new_edge_points.append(ShapelyPoint(test_point))
                if nearest_index['edge_index'] == edge_index \
                    and nearest_index['point_index'] == point_index:
                    new_edge_points.append(ShapelyPoint(point.x, point.y))
            shapely_lines.append(LineString(new_edge_points))
        for test_line in shapely_lines:
            if not test_line.is_valid:
                return self.clone()
        return Line(MultiLineString(shapely_lines), self.spatial_reference, allow_invalid=allow_invalid)

    @property
    def normalized_chord(self):
        """ Returns a numpy.array object representing the 2d normalized vector of the line's endpoints.

        :return: numpy.array of numbers.
        """
        if np.count_nonzero(self.chord) == 0:
            return self.chord
        norm = np.linalg.norm(self.chord)
        return np.true_divide(self.chord, norm)

    def angle_between(self, other_line):
        """ Returns angle in degrees formed between two lines.

        This method measures the angle formed by the vectors formed by the respective endpoints of this Line object and
        another provided Line. This angle is measured from between the two vectors when starting from the same point.

        :param other_line: A Line object to compare angles with.
        :return: Number ranging from 0 to 180.
        """
        dot = np.dot(self.normalized_chord, other_line.normalized_chord)

        # check for -1 (to within 7 decimals) to prevent NaN error
        normalized_dot = round(np.float64(dot).item(), 7)
        return 180 if normalized_dot == float(-1.0) else 0 if normalized_dot == float(1.0) else np.degrees(
            np.arccos(dot))

    def get_line_segment_pairs(self):
        """ Returns every pair of touching line segments in the LineString.

        :return: List of 2-item tuples, each a pair of Line objects.
        """
        if self.point_count < 3:
            return []

        segments = self.split_line_at_vertices()

        # If there are two segments, just return them as the only pair.
        # This avoids the edge case where the second segment is just the first segment in reverse,
        # which will trigger the latter algorithm to return the same pair twice.
        if len(segments) == 2:
            return [segments]

        # All consecutive segments are touching, since they always share a vertex. But we also check
        # on the last segment and first segment, since this may be a loop.
        pairs = list(zip(segments[:-1], segments[1:]))
        if segments[-1].touches(segments[0]):
            pairs.append((segments[-1], segments[0]))
        return pairs

    def split_line_at_vertices(self):
        """ Returns a list of `Line()` objects based on the vertices of the current line.

        :return: A list of `Line()` objects.
        """
        if self._segments:
            return self._segments

        self._segments = []
        points = self.points
        for point_1, point_2 in pairwise(points):
            if point_1.equals(point_2):
                continue
            self._segments.append(Line([point_1, point_2], self.spatial_reference))

        return self._segments

    def union(self, other_line):
        """ Joins this line to the other line.

        :param other_line: The other line object to be joined.
        :return: A new Line object representing the unified line.
        """
        # assert not self.disjoint(other_line)
        union = self.geometry.union(other_line.geometry)
        if isinstance(union, LineString):
            return Line(union, self.spatial_reference)
        if isinstance(union, MultiLineString):
            union = linemerge(union)
            return Line(union, self.spatial_reference)
        return src.oogeo.objects.polygon.Polygon(union, self.spatial_reference)

    def split(self, point):
        """ Splits the line at the indicated point.

        :param point: A `Point()` object representing the split location.
        :return: A tuple containing two `Line()` objects representing the split lines.
        """

        # find the nearest points to the specified point
        nearest_index = None
        for edge_index, edge in enumerate(self.geometry.geoms):
            for point_index in range(len(edge.coords)):
                # if this is the last vertex, associate it with the first one to make an edge
                if point_index == len(edge.coords) - 1:
                    vertex_pair = [src.oogeo.objects.point.Point(edge.coords[point_index], self.spatial_reference),
                                   src.oogeo.objects.point.Point(edge.coords[0], self.spatial_reference)]
                else:
                    vertex_pair = [src.oogeo.objects.point.Point(coords, self.spatial_reference) for coords in
                                   islice(edge.coords, point_index, point_index + 2)]
                if vertex_pair[0].equals(vertex_pair[1]):
                    continue
                test_line = Line(vertex_pair, self.spatial_reference)
                test_index = {
                    'edge_index': edge_index,
                    'point_index': point_index,
                    'distance': point.distance_to(test_line)
                }
                if not nearest_index or test_index['distance'] < nearest_index['distance']:
                    nearest_index = test_index

        # create the new geometries
        shapely_lines = []
        return_line_1 = None
        return_line_2 = None
        for edge_index, edge in enumerate(self.geometry.geoms):
            new_edge_points = []
            for point_index, test_point in enumerate(edge.coords):
                if len(new_edge_points) == 0:
                    new_edge_points.append(ShapelyPoint(test_point))
                else:
                    shapely_test_point = ShapelyPoint(test_point)
                    if not new_edge_points[-1].equals(shapely_test_point):
                        new_edge_points.append(shapely_test_point)
                if nearest_index['edge_index'] == edge_index \
                    and nearest_index['point_index'] == point_index:
                    # split the line here
                    if not new_edge_points[-1].equals(point.geometry):
                        new_edge_points.append(point.geometry)
                    if len(new_edge_points) > 1:
                        shapely_lines.append(LineString(new_edge_points))
                        return_line_1 = Line(MultiLineString(shapely_lines), self.spatial_reference)
                        shapely_lines = []
                        new_edge_points = [point.geometry]
            if len(new_edge_points) > 1:
                shapely_lines.append(LineString(new_edge_points))
        if len(shapely_lines) > 0:
            return_line_2 = Line(MultiLineString(shapely_lines), self.spatial_reference)
        if not return_line_1:
            return return_line_2, None
        return return_line_1, return_line_2

    def shares_a_segment(self, other_line):
        """ Determines if two line objects share a line segment.

        :param other_line: The other line object to be analyzed.
        :return: True of the two edges have a line segment in common, otherwise False.
        """

        # test each vertex to see if two consecutive vertices in the source line are
        # very close to the other line - this would indicate a shared edge
        if has_consecutive_near_points(self, other_line) or has_consecutive_near_points(other_line, self):
            return True

        # if the two lines to not touch at all, return False
        if self.disjoint(other_line):
            return False

        # simplest case - lines share a segment if the lines overlap,
        # if one line either contains or is within the other line, or
        # if the lines are equal
        if self.contains(other_line) or self.overlaps(other_line) or self.within(other_line) or self.equals(other_line):
            return True

        # merge the vertices of the lines together and re-attempt. This uses
        # the "weld point" technique that has been fairly successful correcting
        # polygon cuts.
        merged_self, merged_other = MultipointBase.merge_vertices(self, other_line)
        if merged_self.contains(other_line) or merged_self.overlaps(other_line) or \
                merged_self.within(other_line) or merged_self.equals(other_line):
            return True

        # create small buffers around the edges in order to give them a little
        # thickness, and test with those. This technique has been successful
        # when performing similar queries using spatial SQL (see the
        # `Layer.select_layer_by_location()` method to see the related spatial query)
        buffered_self = self.buffer(0.001, end_style='SQUARE', join_style='MITRE')
        buffered_other = other_line.buffer(0.001, end_style='SQUARE', join_style='MITRE')
        if other_line.overlaps(buffered_self) or other_line.within(buffered_self, exact=True) or \
                self.overlaps(buffered_other) or self.within(buffered_other, exact=True):
            return True

        # test using the intersection of the two lines. This is a different
        # approach for determining if lines share a segment. In this case
        # we expect the intersection of lines that share a segment to
        # be a line object, whereas other relationships (crosses, touches, etc.)
        # will return point objects.
        intersection = self.intersection(other_line)
        if isinstance(intersection, Line):
            return True

        return False

    def is_concave(self, tolerance=135):
        """ Determines if the line is concave within a certain tolerance.

        Concavity in this case is defined as the angle between the first vertex and the last vertex
        exceeds the angle formed by the first and second vertices by more than the specified tolerance.

        :return: `True` if the line is concave within the specified tolerance, otherwise `False`.
        """
        concavity = abs(self.first_angle - self.endpoints_angle)
        return tolerance < concavity < 360 - tolerance

    def snap_to_vertex(self, other_geometry, tolerance, allow_invalid=False):
        """ Snaps the vertices of this geometry to the vertices of the provided geometry. Only points
        that fall within the specified tolerance will be snapped.

        :param other_geometry: The geometry to use as a snapping target.
        :param tolerance: The tolerance to use for snapping.
        :return: A copy of the original geometry with snapped vertices.
        """
        return_geometry = self._snap_(other_geometry, tolerance)
        return_line = Line(return_geometry, self.spatial_reference, allow_invalid=allow_invalid)
        if return_line.is_valid or allow_invalid:
            return return_line
        return self.clone()

    # def snap_to_edge(self, other_geometry, tolerance):
    #     """ Snaps the vertices of this geometry to the vertices of the provided geometry. Only points
    #     that fall within the specified tolerance will be snapped.
    #
    #     :param other_geometry: The geometry to use as a snapping target.
    #     :param tolerance: The tolerance to use for snapping.
    #     :return: A copy of the original geometry with snapped vertices.
    #     """
    #     return_geometry = self.clone()
    #     for point in self.points:
    #         snap_point = other_geometry.snap_point(point)
    #         return_geometry = return_geometry.snap_to_vertex(snap_point, tolerance=tolerance)
    #
    #     return return_geometry

    def snap_to_edge(self, other_geometry, tolerance, allow_invalid=False, densify=False):
        """ Snaps the vertices of this geometry to the vertices of the provided geometry. Only points
        that fall within the specified tolerance will be snapped.

        :param other_geometry: The geometry to use as a snapping target.
        :param tolerance: The tolerance to use for snapping.
        :return: A copy of the original geometry with snapped vertices.
        """
        return_geometry = self.clone()
        other_edge = other_geometry.boundary if isinstance(other_geometry, src.oogeo.objects.polygon.Polygon) else other_geometry
        for point in self.points:
            if other_edge.distance_to(point) > tolerance:
                continue
            snap_point = other_geometry.snap_point(point)
            return_geometry = return_geometry.replace_point(point, snap_point, allow_invalid=allow_invalid)

        if densify:
            for other_point in other_edge:
                if other_point.distance_to(return_geometry) > tolerance or return_geometry.contains_point(other_point):
                    continue
                return_geometry = return_geometry.add_point(other_point)

        return return_geometry

    def to_singlepart(self):
        """ Separates multi-part geometries into single parts.

        :return: A list of geometries representing the parts of the multi-part geometry.
        """

        if not self.is_multipart:
            return [self]
        parts = [part for part in self.geometry.geoms]
        merged_parts = linemerge(parts)
        if isinstance(merged_parts, LineString):
            return [Line(merged_parts, self.spatial_reference)]
        return [Line(part, self.spatial_reference) for part in merged_parts.geoms]

    @staticmethod
    def is_panhandle(first_line, middle_line, last_line, max_panhandle_width=15):
        """ Compares 3 lines together to see if they form a panhandle shape.

        :param first_line: The first line to be compared.
        :param middle_line: The second line to be compared.
        :param last_line: The third line to be compared.
        :return: True if the lines form the shape of a panhandle, otherwise False.
        """
        distance_1 = first_line.midpoint.distance_to(last_line)
        distance_2 = last_line.midpoint.distance_to(first_line)
        panhandle_width = distance_1 if distance_1 < distance_2 else distance_2
        return (panhandle_width <= max_panhandle_width
                and first_line.length > 10 and last_line.length > 10
                and 170 <= (first_line.angle_between(middle_line) + middle_line.angle_between(last_line)) <= 190
                and first_line.turn_direction(middle_line) == middle_line.turn_direction(last_line))

    @staticmethod
    def unify_lines(lines):
        """ Loops through a collection of Line objects and unions them if they touch.

        :param lines: A list of lines to be unified.
        :return: A list of unified lines.
        """

        # if only one line is provided just return the input lines
        if len(lines) == 1:
            return lines

        try:
            unmerged_lines = []
            for multiline in lines:
                for line in multiline.geometry.geoms:
                    unmerged_lines.append(line)
            merged_lines = linemerge(unmerged_lines)

            merged_lines = upgrade_shapely_geometry(merged_lines)
            if not merged_lines:
                # logging.warning(' Cannot unify line geometries. merged_lines is None.')
                return None
            # remove zero-length lines and repair remaining ones
            return [Line(linegeom, lines[0].spatial_reference) for linegeom in merged_lines.geoms if linegeom.length > 0]

        except Exception as e:
            # logging.warning('  Cannot unify invalid line geometries. Exception: %s', e)
            return None

    def clone(self):
        """ Creates a copy of the geometry.

        :return: A copy of the geometry.
        """
        return Line(MultiLineString(self.geometry), self.spatial_reference, allow_invalid=self.allow_invalid)

    def replace_point(self, original_point, new_point, allow_invalid=False):
        """ Will replace the specified vertex with the new vertex. If the new vertex already exists in the
        geometry, the original vertex will be removed and no new vertex will be added.

        :param original_point: The original vertex to be replaced.
        :param new_point: The new vertex.
        :return: A geometry object with updated vertices.
        """

        # loop through the geometry's points and add replace the point
        new_vertices = []
        for point in self.points:
            if point.equals(original_point):
                new_vertices.append(new_point)
            else:
                new_vertices.append(point.clone())

        test_line = Line(new_vertices, self.spatial_reference, allow_invalid=True)
        if test_line.is_valid or allow_invalid:
            return test_line
        return self.clone()



"""
Common functions for parsing well-known string formats.
"""

import re

from pint import UnitRegistry
from shapely.geometry import Polygon as ShapelyPolygon, MultiPolygon as ShapelyMultiPolygon, \
    LineString as ShapelyLineString, MultiLineString as ShapelyMultiLineString
from shapely.wkt import loads as wkt_to_geom

CONVERSION_UNITS = [
    'MILLIMETERS',
    'CENTIMETERS',
    'METERS',
    'KILOMETERS',
    'INCHES',
    'FEET',
    'MILES',
]


def string_to_units(measurement_string, default_units):
    """ Parses a measurement string and converts it to map units. Currently only supports feet and meters,
    but could be expanded in the future.

    :param measurement_string: The measurement string (i.e. '100 FEET')
    :param default_units: The units of the base map (i.e. 'FEET')
    :return: A decimal representing the parsed measurement in map units.
    """
    measurements = measurement_string.split(' ')
    value = float(measurements[0])
    measure_units = measurements[1].upper()

    if 'FOOT' in measure_units or 'FEET' in measure_units:
        measure_units = 'FEET'
    elif 'METER' in measure_units:
        measure_units = 'METERS'

    # if the measure and map units match, return the unadjusted value
    if measure_units == default_units:
        return value

    # handle feet to meters
    if measure_units == 'FEET':
        return value * 0.3048

    # handle meters to feet
    return value * 3.28084

    # TODO: convert to use measurements library (code below)
    # # handle meters to feet
    # if measure_units == 'FEET':
    #     distance = Distance(ft=value)
    #     return distance.m
    #
    # # handle meters to feet
    # distance = Distance(m=value)
    # return distance.ft


def parse_expression(table, field, expression):
    """ Parses esri-style query strings and returns a string that can be injected into SQL.

    :param table: The `Table()` object being updated.
    :param field: The field in the specified table to be updated.
    :param expression: The expression to be evaluated.
    :return: A string that can replace the expression in an SQL statement.
    """

    # return NULL if expression is None
    if expression is None or (isinstance(expression, str) and expression.upper() == 'NULL'):
        return 'NULL'

    # if the expression is not a string, simply return the value
    if not isinstance(expression, str):
        return expression

    # replace field names (bounded by exclamation marks) with SQL-formatted field names
    if expression.count('!') > 0:
        # check for not operator
        if not (expression.count('!=') == 1 and expression.count('!') == 1):
            # check for special case expressions involving shape or oid
            if '!shape.area@squarefeet!' in expression.lower():
                return 'ST_Area("{}")'.format(table.shape)

            # parse out the field name and replace it
            regex_return = re.search('!(.+?)!', expression)
            if regex_return:
                field_name = regex_return.group(1)
                expression = expression.replace('!{}!'.format(field_name), '"{}"'.format(field_name))
    elif 'VARCHAR' in table.get_field_type(field) \
            and not expression.upper().startswith('CASE WHEN') \
            and not expression.upper().startswith('CONCAT') \
            and not expression.upper().startswith('TRIM'):
        # first strip off any surrounding single quotes
        expression = expression.strip()
        expression = expression.strip("'")

        # now escape any unescaped quotes in the string
        if expression.count("'") > 0 and expression.count("''") == 0:
            expression = expression.replace("'", "''")

        # now replace the leading and trailing quotes
        expression = "'{}'".format(expression)

    return expression


def upgrade_shapely_geometry(geometry):
    """

    :param geometry:
    :return:
    """

    rounded_geometry = wkt_to_geom(geometry.wkt)

    return None if not rounded_geometry or rounded_geometry.is_empty \
        else ShapelyMultiPolygon([rounded_geometry]) if isinstance(rounded_geometry, ShapelyPolygon) \
        else ShapelyMultiLineString([rounded_geometry]) if isinstance(rounded_geometry, ShapelyLineString) else rounded_geometry

    ## NOTE: Code below contains an early attempt to clean geometries coming out of Geopandas
    # # if the geometry is multi-part, loop through it's parts and filter out invalid pieces
    # # due to issue w/ geopandas. See https://github.com/geopandas/geopandas/issues/756 for more info.
    # if isinstance(geometry, ShapelyMultiPolygon):
    #     test_polygon = wkt_to_geom(geometry.wkt)
    #     if not test_polygon.is_valid:
    #         clean_polygons = list(filter(lambda x: x.is_valid, test_polygon.geoms))
    #         geometry = ShapelyMultiPolygon(clean_polygons)
    # if isinstance(geometry, ShapelyMultiLineString):
    #     test_line = wkt_to_geom(geometry.wkt)
    #     if not test_line.is_valid:
    #         clean_lines = list(filter(lambda x: x.is_valid, test_line.geoms))
    #         geometry = ShapelyMultiLineString(clean_lines)
    #
    # return ShapelyMultiPolygon([geometry]) if isinstance(geometry, ShapelyPolygon) \
    #     else ShapelyMultiLineString([geometry]) if isinstance(geometry, ShapelyLineString) else geometry


def get_conversion_factor(from_units, to_units):
    """ Returns the conversion factor for converting numbers from one unit to another.

    :param from_units: A string representing the units to be converted from. Valid values are
        "INCHES", "FEET", "MILES", "MILLIMETERS", "CENTIMETERS", "METERS", "KILOMETERS".
    :param to_units: A string representing the units to be converted to. Valid values are
        "INCHES", "FEET", "MILES", "MILLIMETERS", "CENTIMETERS", "METERS", "KILOMETERS".
    :return: A number representing the conversion factor between the from and to units. This
        number can be used as a multiplier to convert any value between the specified units.
    """
    # insure that the units are upper-case
    from_units = from_units.upper()
    to_units = to_units.upper()

    # establish a unit registry in the "from" units
    ureg = UnitRegistry()
    from_reg = None
    if from_units in ['IN', 'INCH', 'INCHES']:
        from_reg = 1 * ureg.inch
    elif from_units in ['FT', 'FOOT', 'FEET']:
        from_reg = 1 * ureg.foot
    elif from_units in ['MI', 'MILE', 'MILES']:
        from_reg = 1 * ureg.mile
    elif from_units in ['MM', 'MILLIMETER', 'MILLIMETERS']:
        from_reg = 1 * ureg.millimeter
    elif from_units in ['CM', 'CENTIMETER', 'CENTIMETERS']:
        from_reg = 1 * ureg.centimeter
    elif from_units in ['M', 'METER', 'METERS']:
        from_reg = 1 * ureg.meter
    elif from_units in ['KM', 'KILOMETER', 'KILOMETERS']:
        from_reg = 1 * ureg.kilometer

    assert from_reg, 'From unit {} not found'.format(from_units)

    # return the conversion units or raise an error
    # if the to units are not specified correctly
    if to_units in ['IN', 'INCH', 'INCHES']:
        return from_reg.to(ureg.inch).magnitude
    elif to_units in ['FT', 'FOOT', 'FEET']:
        return from_reg.to(ureg.foot).magnitude
    elif to_units in ['MI', 'MILE', 'MILES']:
        return from_reg.to(ureg.mile).magnitude
    elif to_units in ['MM', 'MILLIMETER', 'MILLIMETERS']:
        return from_reg.to(ureg.millimeter).magnitude
    elif to_units in ['CM', 'CENTIMETER', 'CENTIMETERS']:
        return from_reg.to(ureg.centimeter).magnitude
    elif to_units in ['M', 'METER', 'METERS']:
        return from_reg.to(ureg.meter).magnitude
    elif to_units in ['KM', 'KILOMETER', 'KILOMETERS']:
        return from_reg.to(ureg.kilometer).magnitude
    raise ValueError('To unit {} not found'.format(to_units))



"""
Contains functions common to all tools.
"""

import logging


def configure_logger(
    log_level='INFO',
    log_file=None,
    log_name='main'):
    """
    Configures the logger based on inputs from the user at the command line.

    :param log_level: The log level to be used by the application. The default value is logging.INFO
    :param log_file: The fully qualified path to a logging file where the logging output will be written.
    """

    # logging format constants
    log_format = '%(asctime)s %(message)s'
    date_format = '%m/%d/%Y %I:%M:%S %p'

    level = getattr(logging, log_level.upper()) if isinstance(log_level, str) else log_level
    # configure logging
    logger = logging.getLogger(log_name)
    logger.setLevel(level)
    formatter = logging.Formatter(fmt=log_format, datefmt=date_format)
    if log_file:
        handler = logging.FileHandler(log_file)
    else:
        handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.propagate = False
    return logger

def determine_oid_range(arguments):
    """

    """
    return '{0}-{0}'.format(arguments.oid) if arguments.oid else arguments.oid_range

def parse_args_and_configure_logger(argument_parser):
    """

    """
    arguments = argument_parser.parse_args()

    # configure logging
    logger = configure_logger(
        log_level=arguments.log_level.upper(),
        log_file=arguments.log_file)
    return arguments, logger

def add_input_and_output_parameters(argument_parser):
    """ Adds standard input and output arguments to the argument parser.

    :param argument_parser: The argument parser object to be modified.
    :return: None
    """
    argument_parser.add_argument('-in', '--inputs',
        metavar='Inputs connection',
        dest='inputs_connection',
        default='Inputs',
        help='The connection name from the connections.ini file for the inputs geodatabase.')
    argument_parser.add_argument('-out', '--outputs',
        metavar='Outputs connection',
        dest='outputs_connection',
        default='Processing',
        help='The connection name from the connections.ini file for the outputs geodatabase.')


def add_oid_parameters(argument_parser):
    """ Adds parameters for collecting object ids to process.

    :param argument_parser: The argument parser object to be modified.
    :return: None
    """
    argument_parser.add_argument('-oids', '--oid_range',
        metavar='Physical Block Object Ids',
        dest='oid_range',
        help='A range of physical block object ids to process. The format is a string with a dash separating the lower '
            'and upper boundaries of the range (ex: "100-200"). The range is inclusive and will include the lower '
            'and upper boundary values in the search.')
    argument_parser.add_argument('-oid', '--oid',
        metavar='Physical Block Object Id',
        dest='oid',
        type=int,
        default=None,
        help='A single physical block object id to process.')

def add_multiprocessing_parameters(argument_parser):
    """ Adds parameters for collecting multiprocessing settings.

    :param argument_parser: The argument parser object to be modified.
    :return: None
    """
    argument_parser.add_argument('-p', '--processes',
        metavar='Processes',
        dest='processes',
        type=int,
        default=1,
        help='The number of processes to use when running the tool. If not specified one process will be used.')
    argument_parser.add_argument('-d', '--database',
        metavar='Database',
        dest='database',
        help='The preconfigured database to use for processing (should only be used internally when multiprocessing).')
    argument_parser.add_argument('-eo', '--errors_only',
        dest='errors_only',
        action='store_true',
        help='Only creates block outputs for blocks that appear to be missing or corrupt. This option can be used to '
             'restart failed processes when necessary.')


def add_change_input_parameters(argument_parser):
    """ Adds parameters for connecting to and using the changes geodatabase for changes-only processing.

    :param argument_parser: The argument parser object to be modified.
    :return: None
    """
    argument_parser.add_argument('-cin', '--change_inputs',
        metavar='Change inputs connection',
        dest='change_connection',
        default='Differences',
        help='The connection name from the connections.ini file for the pluto differences geodatabase.')
    argument_parser.add_argument('-co', '--changes_only',
        dest='changes_only',
        action='store_true',
        help='Only processes blocks found in the changed blocks table.')


def add_database_name_parameters(argument_parser):
    """ Adds -id (input database) and -od (output database) to the provided argument parser.

    :param argument_parser: The argument parser object to be modified.
    :return: None
    """
    argument_parser.add_argument('-id', '--input_database',
        metavar='Input database name',
        dest='input_database',
        default=None,
        help='If specified, the provided database name will override the database name derived from the connections.ini file for the inputs connection (useful for transferring working data).')
    argument_parser.add_argument('-od', '--output_database',
        metavar='Output database name',
        dest='output_database',
        default=None,
        help='If specified, the provided database name will override the database name derived from the connections.ini file for the outputs connection (useful for transferring working data).')


def add_locale_parameters(argument_parser):
    """ Adds -id (input database) and -od (output database) to the provided argument parser.

    :param argument_parser: The argument parser object to be modified.
    :return: None
    """
    argument_parser.add_argument('locale',
        help='The region to process. Valid values are "NYC", "MIA", or "DC"',
        choices=['NYC', 'MIA', 'DC'])

def add_username_password_parameters(argument_parser):
    """ Adds username and password parameters to the provided argument parser.

    :param argument_parser: The argument parser object to be modified.
    :return: None
    """
    argument_parser.add_argument('-iun', '--input_user_name',
        metavar='Input user name',
        dest='input_user',
        default=None,
        help='If specified, the provided user name will override the user name derived from the connections.ini file for the inputs connection.')
    argument_parser.add_argument('-ipw', '--input_password',
        metavar='Input password',
        dest='input_password',
        default=None,
        help='If specified, the provided password will override the password derived from the connections.ini file for the inputs connection.')
    argument_parser.add_argument('-oun', '--output_user_name',
        metavar='Output user name',
        dest='output_user',
        default=None,
        help='If specified, the provided user name will override the user name derived from the connections.ini file for the outputs connection.')
    argument_parser.add_argument('-opw', '--output_password',
        metavar='Output password',
        dest='output_password',
        default=None,
        help='If specified, the provided password will override the password derived from the connections.ini file for the outputs connection.')



def add_logging_parameters(argument_parser, log_level='INFO'):
    """ Adds log level and log file inputs to the supplied argument parser.

    :param argument_parser: The argument_parser object to add arguments to.
    :param log_level: Configures the default log level for the tool. The default is 'INFO'.
    """
    argument_parser.add_argument('-ll', '--log_level',
        dest='log_level',
        help='Sets the logging level for the application.',
        default=log_level,
        metavar='Log Level',
        choices=['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'])
    argument_parser.add_argument('-lf', '--log_file',
        metavar='Log File',
        dest='log_file',
        help='Sets the path to the log file.')



"""
Common functions for creating database queries
"""

from shapely.wkb import loads as wkb_to_geom
from shapely.geometry import Polygon as ShapelyPolygon, LineString, MultiPolygon, MultiLineString

import src.oogeo.objects.line
import src.oogeo.objects.point
import src.oogeo.objects.polygon

from src.oogeo.functions.conversions import upgrade_shapely_geometry
from src.oogeo.objects.polygon import correct_invalid_polygon
from src.oogeo.objects.line import correct_invalid_line

from geoalchemy2 import Geometry, WKBElement


def geodataframe_to_layer(geodataframe, workspace, output_schema, output_layer_name, oid_field, shape_field):
    """ Converts a geodataframe to a geodatabase layer.

    :param geodataframe: The geodataframe to be stored in the database.
    :param workspace: The workspace containing the connection to the database where the geodataframe should be stored.
    :param output_layer_name: The name of the table to be created in the geodatabase.
    :param oid_field: The field representing the primary key for the table.
    :param shape_field: The field representing the shape field for the table.
    :return: A `Layer()` object representing the geodatabase table that was created.
    """

    # get the spatial reference in epsg format
    epsg = int(geodataframe.crs.to_epsg())

    # upgrade geometries to multis
    geodataframe.geometry = geodataframe.geometry.apply(upgrade_shapely_geometry)

    # get the layer's geometry type
    geometry_type = str(geodataframe.geom_type.tolist()[0]).upper()

    # correct invalid geometries if necessary
    if geometry_type == 'MULTIPOLYGON':
        geodataframe.geometry = geodataframe.geometry.apply(correct_invalid_polygon)
        geodataframe.drop(geodataframe[geodataframe.geometry.isnull()].index, inplace=True)

    elif geometry_type == 'MULTILINESTRING':
        geodataframe.geometry = geodataframe.geometry.apply(correct_invalid_line)

        # delete rows with null geometries
        geodataframe.drop(geodataframe[geodataframe.geometry.isnull()].index, inplace=True)

    # translate the geometries to geoalchemy (makes them storable in the db)
    geodataframe['temp_shape'] = geodataframe.geometry.apply(lambda x: WKBElement(x.wkb, srid=epsg))

    # delete the current geometry field and rename the new one properly
    geodataframe.drop(geodataframe.geometry.name, axis=1, inplace=True)
    geodataframe = geodataframe.rename(columns={'temp_shape': shape_field})

    # drop the edge_buffers column if it exists
    if 'edge_buffers' in list(geodataframe.columns.values):
        geodataframe.drop('edge_buffers', 1, inplace=True)
    if 'EDGE_BUFFERS' in list(geodataframe.columns.values):
        geodataframe.drop('EDGE_BUFFERS', 1, inplace=True)

    # drop the centroid column if it exists
    if 'centroid' in list(geodataframe.columns.values):
        geodataframe.drop('centroid', 1, inplace=True)
    if 'CENTROID' in list(geodataframe.columns.values):
        geodataframe.drop('CENTROID', 1, inplace=True)

    # drop the midpoint column if it exists
    if 'midpoint' in list(geodataframe.columns.values):
        geodataframe.drop('midpoint', 1, inplace=True)
    if 'MIDPOINT' in list(geodataframe.columns.values):
        geodataframe.drop('MIDPOINT', 1, inplace=True)

    # store the layer in the db
    workspace.create_schema(output_schema)
    geodataframe.to_sql(output_layer_name, workspace.alchemy_engine, if_exists='fail', index_label=oid_field, schema=output_schema,
        dtype={shape_field: Geometry(geometry_type, srid=epsg)})

    # set the object id as the primary key
    workspace.execute_simple_sql('ALTER TABLE "{}"."{}" ADD PRIMARY KEY ("{}");'.format(output_schema, output_layer_name, oid_field),
        use_transaction=False)

    # create a sequence for the table
    workspace.create_sequence(output_layer_name, oid_field, schema=output_schema)

    return_layer = oogeo.objects.layer.Layer(output_layer_name, workspace, schema=output_schema)
    return return_layer


def create_in_clause(inclusion_list, use_quotes=False, field_name=None, not_operator=False):
    """
    Creates an SQL "IN" clause base on the data in the provided list.

    :param inclusion_list: The list of ids to convert into an IN clause.
    :param use_quotes: If True, single-quotes will be added around each value
    :param field_name: A field name to append to the query.
    :param not_operator: Adds the NOT operator to reverse the results of the in clause.
    :return: A string with the complete in clause.
    """
    if len(inclusion_list) == 0:
        inclusion_list.add(-1)
    quote_mark = "'" if use_quotes else ''
    not_operator = ' NOT' if not_operator else ''
    return_string = ''
    # get the list of values in a comma-separated list
    for item in inclusion_list:
        # append a comma if the list has already been started
        if not return_string == '':
            return_string = '{},'.format(return_string)
        # append current record
        return_string = '{}{}{}{}'.format(return_string, quote_mark, item, quote_mark)

    # create the final output and return
    field_name = '' if field_name is None else '"{}"'.format(field_name) if '.' not in field_name else field_name
    return '{}{} IN ({})'.format(field_name, not_operator, return_string)


def prepare_query(table, field_names, query_string, sql_clause, use_raw_fields=False):
    """ Prepares a complete SQL statement based on the inputs provided.

    :param table: The table that the query will be based on.
    :param field_names: The list of field names to be included in the query.
    :param query_string: The FROM clause to apply to the query.
    :param sql_clause: A tuple representing the prefix and postfix to be applied to the query.
        Typically the prefix is used to apply a COUNT or DISTINCT clause, and the postfix is
        used to apply an ORDER BY clause.
    :return: A complete SQL statement based on the provided inputs.
    """
    query = {}

    # Build up the correct WHERE clause. If the function call did not have one as a parameter,
    # we default to using the current layer selection (both attribute and location). If neither
    # exists, we omit the WHERE clause entirely.
    if not query_string:
        query['from_clause'] = f'"{table.schema}"."{table.table_name}"'
        query['where_clause'] = ""
    elif isinstance(query_string, int):
        query['from_clause'] = f'"{table.schema}"."{table.table_name}"'
        query['where_clause'] = 'WHERE "{}" = {}'.format(table.oid, query_string)
    elif isinstance(query_string, str):
        query['from_clause'] = table.from_clause
        query['where_clause'] = "WHERE {}".format(query_string)
    else:
        raise ValueError('Need valid query_string')

    # prepare field names
    if isinstance(field_names, str):
        field_names = [field_names]

    if '*' in field_names:
        field_names = [table.shape, table.oid] + table.field_names \
            if hasattr(table, 'shape') else [table.oid] + table.field_names

    table_oid_sql = '"{}" AS table_oid'.format(table.oid) if not use_raw_fields else '"{}"'.format(table.oid)
    shape_field_names = [name.lower() for name in table.get_geometry_field_names(include_hidden=True)] \
        if hasattr(table, 'get_geometry_field_names') else []
    column_names = {
        table_oid_sql if name.lower() == 'oid@' or name.lower() == table.oid.lower() else
        'ST_AsEWKB("{0}") as "{0}"'.format(table.shape) if not use_raw_fields and ('shape@' in name.lower() or (hasattr(table, 'shape') and name.lower() == table.shape.lower())) else
        'ST_AsEWKB("{0}") as "{0}"'.format(name) if not use_raw_fields and name.lower() in shape_field_names else
        '"{}"'.format(name)
        for name in field_names
    }
    if table_oid_sql not in column_names:
        column_names.add(table_oid_sql)

    query['columns'] = ", ".join(column_names)

    # Build up prefix and postfix clauses, if provided.
    if (sql_clause is None or sql_clause[0] is None) and table.from_clause != f'"{table.schema}"."{table.table_name}"':
        # this is a spatial join - we need to add a distinct clause here
        if not sql_clause:
            sql_clause = ('DISTINCT', None)
        elif not sql_clause[0] and (not sql_clause[1] or not sql_clause[1].upper().startswith('ORDER BY')):
            sql_clause = ('DISTINCT', sql_clause[1])

    query['prefix'] = sql_clause[0] if sql_clause and sql_clause[0] else ""
    query['postfix'] = sql_clause[1] if sql_clause and sql_clause[1] else ""

    return "SELECT {prefix} {columns} FROM {from_clause} {where_clause} {postfix}".format(**query)


def convert_feature_to_row(source_table, feature, field_names, allow_invalid):
    """ Converts the provided feature into a list representing the row.

    :param source_table: The source table that the features came from.
    :param feature: The feature to be converted.
    :param field_names: The names of the fields in the feature.
    :return: A List representing the values in the row, ordered by the provided list of field names.
    """
    row = {}

    # iterate through field names and assemble row
    for field_name in field_names:
        lcase_field_name = field_name.lower()
        if 'oid@' in lcase_field_name or source_table.oid == field_name:
            row[field_name] = feature['table_oid']
            continue
        geometry_fields = [name.lower() for name in source_table.get_geometry_field_names(include_hidden=True)] if hasattr(source_table, 'get_geometry_field_names') else []
        if lcase_field_name not in ['shape@', 'shape@wkt', 'shape@xy', 'shape@length'] + geometry_fields:
            # check to see if the row is a boolean
            if source_table.get_field(field_name).type == 'BOOLEAN':
                row[field_name] = bool(feature[field_name])
            else:
                row[field_name] = feature[field_name]
            continue

        # Cache the geometry
        shape_field = field_name if lcase_field_name in geometry_fields else source_table.shape
        load_invalid = True if lcase_field_name in ['edge_buffers', 'centroid', 'midpoint'] else allow_invalid
        shapely_geometry = wkb_to_geom(feature[shape_field].tobytes())
        try:
            if isinstance(shapely_geometry, ShapelyPolygon) or isinstance(shapely_geometry, MultiPolygon):
                geometry = src.oogeo.objects.polygon.Polygon(shapely_geometry, source_table.spatial_reference, allow_invalid=load_invalid)
            elif isinstance(shapely_geometry, LineString) or isinstance(shapely_geometry, MultiLineString):
                geometry = src.oogeo.objects.line.Line(shapely_geometry, source_table.spatial_reference, allow_invalid=load_invalid)
            else:
                geometry = src.oogeo.objects.point.Point(shapely_geometry, source_table.spatial_reference, allow_invalid=load_invalid)
        except ValueError as e:
            raise ValueError('Invalid geometry in layer {}: {}'.format(source_table.table_name, shapely_geometry.wkt))

        # Lookup geometry fields
        if lcase_field_name in ['shape@'] + geometry_fields:
            row[field_name] = geometry
        elif lcase_field_name == 'shape@wkt':
            row[field_name] = geometry.wkt
        elif lcase_field_name == 'shape@length':
            if isinstance(geometry, src.oogeo.objects.point.Point):
                row[field_name] = 0
            else:
                row[field_name] = geometry.length
        elif lcase_field_name == 'shape@xy':
            if isinstance(geometry, src.oogeo.objects.point.Point):
                row[field_name] = geometry
            else:
                row[field_name] = geometry.centroid
        else:
            row[field_name] = None
    return row


def convert_features_to_rows(source_table, result_layer, field_names, allow_invalid):
    """ Converts the features in a query result table to a list of rows

    :param source_table: The table that spawned the query.
    :param result_layer: The result layer returned from the query.
    :param field_names: The field names to extract from the features.
    :return: A list of values representing the rows in the result layer.
    """

    results = []
    for current_feature in result_layer:
        results.append(convert_feature_to_row(source_table, current_feature, field_names, allow_invalid))
    return results


def filter_query_results(selection, current_filtered_oids, unfiltered_results, oid_field, shape_field):
    """ Manually filters query results based on geometry analysis.

    :param selection: A dictionary containing the selection criteria.
    :param current_filtered_oids: The list of currently selected oids from previous queries.
    :param unfiltered_results: A list of unfiltered results from the database.
    :return: A list of object ids representing the items to include in the final result set.
    """
    filtered_oids = current_filtered_oids.copy() \
        if not (selection['selection_type'] == 'SUBSET_SELECTION'
                or selection['selection_type'] == 'SWITCH_SELECTION') else set()

    # iterate through results and filter them manually
    result_oid = -1
    for row in unfiltered_results:
        test_oid = row[oid_field]
        result_geometry = row[shape_field]
        # handle custom geometry searches here based on overlap type
        if selection['overlap_type'] == 'SHARE_A_LINE_SEGMENT_WITH':
            result_lines = result_geometry.to_lines() if isinstance(result_geometry, src.oogeo.objects.polygon.Polygon) else [
                result_geometry]
            # filter results based on overlap type and collect oids
            for geometry in selection['search_objects']:
                geometry_lines = geometry.to_lines() if isinstance(geometry, src.oogeo.objects.polygon.Polygon) else [geometry]
                for result_line in result_lines:
                    for geometry_line in geometry_lines:
                        if result_line.shares_a_segment(geometry_line):
                            result_oid = test_oid
                            break
                    if result_oid >= 0:
                        break
                if result_oid >= 0:
                    break
        elif selection['overlap_type'] == 'BOUNDARY_TOUCHES':
            for geometry in selection['search_objects']:
                if result_geometry.touches(geometry):
                    result_oid = test_oid
                    break
        elif selection['overlap_type'] == 'CROSSED_BY_THE_OUTLINE_OF':
            for geometry in selection['search_objects']:
                if isinstance(result_geometry, src.oogeo.objects.polygon.Polygon):
                    lines = result_geometry.to_lines()
                    for line in lines:
                        if line.crosses(geometry):
                            result_oid = test_oid
                            break
                else:
                    if result_geometry.crosses(geometry):
                        result_oid = test_oid
                        break
        else:
            result_oid = row[oid_field]

        # if we found a result oid, handle it in our list
        if result_oid >= 0:
            if selection['selection_type'] == 'REMOVE_FROM_SELECTION':
                if result_oid in filtered_oids:
                    filtered_oids.remove(result_oid)
            elif selection['selection_type'] == 'SUBSET_SELECTION':
                if result_oid in current_filtered_oids:
                    filtered_oids.add(result_oid)
            else:
                filtered_oids.add(result_oid)

    return filtered_oids


def append_query_string(selection_type, current_query_string, new_query_string):
    """ Appends the new query string to the current string based on selection type.

    :param selection_type: The selection type, which will determine how the queries are joined.
    :param current_query_string: The current query string.
    :param new_query_string: The new query string to be appended to the current string.
    :return: The appended query string.
    """
    if selection_type == 'NEW_SELECTION':
        return new_query_string
    if selection_type == 'ADD_TO_SELECTION':
        return "({}) OR {}".format(current_query_string, new_query_string)
    if selection_type == 'REMOVE_FROM_SELECTION':
        return "({}) AND NOT {}".format(current_query_string, new_query_string)
    if selection_type == 'SUBSET_SELECTION':
        return "({}) AND {}".format(current_query_string, new_query_string)
    # handle final case - 'SWITCH_SELECTION'
    return "NOT ({})".format(current_query_string)


def execute_immediate(selections):
    """ Determines if each query needs to be executed and filtered individually or if a query string can be created.
    Only selection sets that are entirely attribute queries can use an accumulated query string. Selection sets
    that contain geometry queries must be executed individually.

    :param selections: The list of selections in the seleciton set.
    :return: True if each selection should be queried immediately, otherwise False.
    """
    for selection in selections:
        # determine if we can need to run the query now or if it can wait
        # if selection['overlap_type'] in ['SHARE_A_LINE_SEGMENT_WITH', 'BOUNDARY_TOUCHES', 'CROSSED_BY_THE_OUTLINE_OF']:
        if selection['overlap_type'] != 'ATTRIBUTE':
            return True
    return False


def format_query_string(table, query_string):
    """

    :param query_string:
    :return:
    """
    if isinstance(query_string, int):
        return query_string
    shape_field = table.shape if hasattr(table, 'shape') else ''
    query_string = ' {}'.format(query_string)
    for field_name in table.field_names + [table.oid, shape_field]:
        query_string = query_string.replace(' {} '.format(field_name), ' "{}" '.format(field_name))
    query_string = query_string.replace('%', '%%')
    return query_string.strip()


def build_update_clause(table, selections):
    """

    :param table:
    :param selections:
    :return:
    """
    update_clause = ''
    for selection in selections:
        if 'result_query_string' in selection:
            update_clause = selection['result_query_string']
        else:
            update_clause = append_query_string(selection['selection_type'], update_clause, format_query_string(table, selection['update_clause']))
    return 'WHERE {}'.format(update_clause) if update_clause != '' else update_clause


def build_query_string(table, selections, allow_invalid=False):
    """ Builds a query string (WHERE clause) for the provided table based on the list of provided selections.

    :param table: The table to be queried.
    :param selections: The list of dictionaries representing the selections to be made on the table.
    :return: A query string representing the WHERE clause that will return the selected results from the table.
    """

    # loop through selections and build up selection results
    query_string = ''
    from_tables = set()
    filtered_oids = set()
    execute = execute_immediate(selections)
    previous_query_string = None
    for selection in selections:

        # if this portion of the query has already been executed, pull
        # the result query string and move to the next selection object
        if 'result_query_string' in selection:
            previous_query_string = selection['result_query_string']
            query_string = selection['result_query_string']
            filtered_oids = selection['result_object_ids']
            from_tables = set()
            continue

        # begin building up query
        formatted_query = format_query_string(table, selection['query'])
        query_string = append_query_string(selection['selection_type'], query_string, formatted_query) \
            if not execute else formatted_query if not selection['selection_type'] == 'SWITCH_SELECTION' \
            else create_in_clause(filtered_oids, field_name=table.oid, not_operator=True)
        from_tables.update(selection['from_tables'])

        # this is a geometry query - we will need to execute the
        # query and manually filter the results
        if execute:
            sql_clause = ('DISTINCT', None) if selection['overlap_type'] != 'ATTRIBUTE' else (None, None)
            full_query = prepare_query(table, [table.oid, table.shape], query_string, sql_clause)

            # if this is a subset or add to selection, attach the previous results to the query to make it faster
            if previous_query_string:
                full_query = '{} AND ({})'.format(full_query, previous_query_string) \
                    if selection['selection_type'] == 'SUBSET_SELECTION' \
                    else '{} AND NOT ({})'.format(full_query, previous_query_string) \
                    if selection['selection_type'] == 'ADD_TO_SELECTION' else full_query

            # execute the query
            result_layer = table.workspace.execute_simple_sql(full_query)

            # get a list of features to filter
            unfiltered_results = convert_features_to_rows(table, result_layer, [table.oid, table.shape], allow_invalid)

            # filter the features based on their geometries
            filtered_oids = filter_query_results(selection, filtered_oids, unfiltered_results, table.oid, table.shape)

            # replace the current sql statement with an IN clause
            # selecting the oids of the required features, and reset
            # the list of tables (it will be repopulated on the next loop)
            query_string = create_in_clause(filtered_oids, field_name='{}."{}"'.format(table.table_name, table.oid))
            from_tables = set()

            # cache the resulting IN clause and filtered oids for re-use later
            selection['result_query_string'] = query_string
            selection['update_clause'] = query_string
            selection['result_object_ids'] = filtered_oids
            selection['from_tables'] = {f'"{table.schema}"."{table.table_name}"'}
    # return complete query string when complete
    return query_string


def create_insert_row(table, row):
    """ Creates a row that can be inserted into the provided table.

    """
    # map field names to values
    insert_fields = {}
    geometry_columns = [column.lower() for column in
                        table.get_geometry_field_names(include_hidden=True)] if hasattr(table, 'get_geometry_field_names') else []
    for field_name, value in row.items():
        if field_name.lower() in geometry_columns + ['shape@']:
            field_name = table.shape if field_name.lower() == 'shape@' else field_name
            insert_fields[field_name] = value.to_postgis() if hasattr(value, 'to_postgis') else value
        elif field_name.lower() == 'oid@' or field_name.lower() == table.oid.lower():
            continue
        elif field_name in table.field_names:
            insert_fields[field_name] = value
        else:
            raise KeyError('Could not find field {}'.format(field_name))
    return insert_fields

